(** Executable: Builds up the key database from a multi-file database dump.
  dump files are taken from the command-line 
*)
open StdLabels
open MoreLabels
open Printf
open Arg
open Common
module Set = PSet.Set
open Packet
module Keydb = Keydb.Make(struct 
			    let withtxn = false
			    and cache_bytes = !Settings.cache_bytes
			    and pagesize = !Settings.pagesize
			    and dbdir = !Settings.dbdir
			    and dumpdir = !Settings.dumpdir
			  end)

let n = match !Settings.n with 0 -> 1 | x -> x
let fnames = !Settings.anonlist

let rec get_keys_rec nextkey partial = match nextkey () with 
	Some key -> get_keys_rec nextkey (key::partial)
      | None -> partial
	    
let get_keys nextkey = get_keys_rec nextkey []

let timestr sec = 
  sprintf "%.2f min" (sec /. 60.)
    
let rec nsplit n list = match n with
    0 -> ([],list)
  | n -> match list with
	[] -> ([],[])
      | hd::tl -> 
	  let (beginning,ending) = nsplit (n-1) tl in
	  (hd::beginning,ending)

let rec batch_iter ~f n list = 
  match nsplit n list with
      ([],_) -> ()
    | (firstn,rest) -> f firstn; batch_iter ~f n rest

let get_keys_fname fname start = 
  let cin = new Channel.sys_in_channel (open_in fname) in
  protect 
    ~f:(fun () -> 
	  let nextkey = Key.next_of_channel cin in
	  get_keys_rec nextkey start
       )
    ~finally:(fun () -> cin#close)

let get_keys_multi flist = 
  List.fold_left ~f:(fun keys fname -> get_keys_fname fname keys)
    flist ~init:[]

let dbtimer = MTimer.create ()
let timer = MTimer.create ()
let () = 
  if Sys.file_exists !Settings.dbdir then
    ignore (Unix.system (sprintf "rm -r %s" !Settings.dbdir));
  Unix.mkdir !Settings.dbdir 0o700; 

  Keydb.open_dbs ();
  
  protect 
    ~f:(fun () ->
	  batch_iter n fnames 
	  ~f:(fun fnames ->
		MTimer.start timer;
		printf "Loading keys..."; flush stdout;
		let keys = get_keys_multi fnames in
		printf "done\n"; flush stdout;
		MTimer.start dbtimer;
		Keydb.add_keys keys;
		MTimer.stop dbtimer;
		MTimer.stop timer;
		printf "DB time:  %s.  Total time: %s.\n" 
		  (timestr (MTimer.read dbtimer)) 
		  (timestr (MTimer.read timer)); 
		flush stdout;
	     )
       )
    ~finally:(fun () -> Keydb.close_dbs)

