(** Executable: Cleans up various problems that occur in key
  databases. Currently, this includes:
  - Merging all mergeable keys
*)

open StdLabels
open MoreLabels
open Printf
open Arg
open Common
module Set = PSet.Set
module Map = PMap.Map
module Unix = UnixLabels
open Packet
open Bdb

let ( |= ) map key = Map.find key map
let ( |< ) map (key,data) = Map.add ~key ~data map 

module Keydb = Keydb.Make(struct 
			    let withtxn = !Settings.transactions
			    and cache_bytes = !Settings.cache_bytes
			    and pagesize = !Settings.pagesize
			    and dbdir = !Settings.dbdir
			    and dumpdir = !Settings.dumpdir
			  end)

let rec get_dups_rec cursor accum =
  try 
    let (key,data) = Cursor.get cursor Cursor.NEXT_DUP [] in
    get_dups_rec cursor ((key,data)::accum)
  with
      Not_found -> accum

let get_dups cursor = 
  let pairs = get_dups_rec cursor [] in
  match pairs with
      [] -> failwith "get_dups retrieved empty list"
    | (key,data)::tail ->
	let dtail = List.map tail
		      ~f:(fun (tkey,tdata) -> if tkey <> key 
			  then failwith "get_dups retrieved non-duplicate"
			  else tdata
			 )
	in
	(key,data::dtail)
  
let get_keypacket pkey = 
  match pkey with
      KeyMerge.V3 v3key -> v3key.KeyMerge.v3_key
    | KeyMerge.V4 v4key -> v4key.KeyMerge.v4_key

let rec join_by_keypacket map keylist = match keylist with 
  | [] -> Map.fold ~f:(fun ~key ~data list -> !data::list) ~init:[] map
  | key::tl -> 
      let keypacket = get_keypacket key in
      let map = 
	try
	  let keylist_ref = map |= keypacket in
	  keylist_ref := key::!keylist_ref;
	  map
	with
	    Not_found ->
	      map |< (keypacket,ref [key])
      in
      join_by_keypacket map tl

(** Returns two lists of keys: the first list containing keys that could not
  be parsed, the second list containing the parsed versions of the rest of the
  keys *)
let parse_keys keys = 
  List.fold_left keys ~init:([],[])
    ~f:(fun (unparseable,parsed) key ->
	  try 
	    let pkey = KeyMerge.key_to_pkey key in
	    (unparseable,pkey::parsed)
	  with
	      Stream.Failure | Stream.Error _ -> 
		(key::unparseable,parsed)
       )

let unconditional_merge pkey1 pkey2 = 
  match KeyMerge.merge_pkeys pkey1 pkey2 with
      None -> failwith "Attempt to merge keys failed"
    | Some pkey -> pkey

let merge_pkey_list keys = match keys with
    first::rest ->
      let pkey = List.fold_left ~f:unconditional_merge ~init:first rest in
      KeyMerge.flatten pkey
  | [] -> 
      failwith "Attempt to merge empty list of keys"

(** checks if a sorted list has duplicates *)
let rec has_dups list = match list with 
    [] -> false
  | [hd] -> false
  | hd1::hd2::tl -> 
      if hd1 = hd2 then true
      else has_dups (hd2::tl)

let merge_from_hashes hashes = 
  let hashes = List.sort ~cmp:compare hashes in
  if has_dups hashes then 
    signore (perror "Duplicates found in hash list"); 
  let keys = List.map hashes 
	       ~f:(fun hash -> 
		     let key = Keydb.get_by_hash hash in
		     let newhash = KeyHash.hash key in
		     if newhash <> hash then
		       signore 
			 (perror "Key hashes don't match up:\n\t%s\n\t%s" 
			    (KeyHash.hexify hash) (KeyHash.hexify newhash));
		     key
		  )
  in
  let (unparseable,pkeys) = parse_keys keys in
  if List.length unparseable > 0 then
    signore (perror "%d unparseable keys found" (List.length unparseable));
  let keylists = join_by_keypacket Map.empty pkeys in
  List.iter keylists
    ~f:(fun pkeys -> 
	  if List.length pkeys > 1 then 
	    let merged_key = merge_pkey_list pkeys in
	    signore (perror "Replacing %d keys with single merged key" 
		       (List.length pkeys));
	    List.iter pkeys 
	      ~f:(fun pkey -> 
		    let hash = KeyHash.hexify 
				 (KeyHash.hash (KeyMerge.flatten pkey)) in
		    signore (perror "Key to be deleted: %s" hash) 
		 );
	    let hash = KeyHash.hexify (KeyHash.hash merged_key) in
	    signore (perror "Key to add: %s" hash)
	  else
	    signore (perror "Only 1 key to merge, doing nothing.")
       )

let ctr = ref 0 
let tick () = 
  incr ctr;
  if !ctr mod 5000 = 0 then
    signore (perror "%d thousand keyids processed" (!ctr/1000))

let run () = 
  Keydb.open_dbs ();
  let dbs = Keydb.get_dbs () in
  let c = Cursor.create dbs.Keydb.keyid in
  
  let (keyid,hash) = Cursor.get c Cursor.FIRST [] in
  
  let finished = ref false 
  and first_keyid = ref keyid
  and first_hash = ref hash 
  in
  while not !finished do
    tick ();
    if Cursor.count c > 1 then (
      let (keyid,hashes) = get_dups c in
      if keyid <> !first_keyid then failwith "Failure retrieving duplicates";
      let hashes = !first_hash::hashes in
      merge_from_hashes hashes
    );
    try
      let (keyid,hash) = Cursor.get c Cursor.NEXT [] in
      first_keyid := keyid;
      first_hash := hash
    with
	Not_found -> finished := true
  done


let () = 
  if not !Sys.interactive then run ()
