(** Client side of set-reconciliation algorithm *)
open StdLabels
open MoreLabels
module Unix=UnixLabels
open Common

open Printf
open ReconMessages

module Set = PSet.Set
module Map = PMap.Map
module PTree = PrefixTree
(* module ZZp = RMisc.ZZp *)

exception Bug of string

(***************************************************************)
(*  Diagnostic Timers  *****************************************)
(***************************************************************)

let flushcount = ref 0

let timer = MTimer.create ()
let tstart () = 
  MTimer.start timer
let tstop accum =
  MTimer.stop timer;
  accum := !accum +. MTimer.read_ms timer

let get_flushcount () = !flushcount

(***************************************************************)
(***************************************************************)
(***************************************************************)

(* NOTE:  the code below could be improved by eliminating send_available.  
   read_available would simply do the sending of the new messages, updating a
   second requestQ  as appropriate for any messages that will be responded
   to.  This approach should also be better from a performance perspective,
   in addition to being somewhat simpler. 
*)


(** The client code is based on a simple, event-based structure.  There are
  two queues, a requestQ which tells you the requests you need to send, and
  the bottomQ, which tells you the responses you're waiting to handle.
 
  Both the requestQ and bottomQ contain simple (node,key) pairs, where
  node is the relevant PTree node, and key is the prefix-key of that node.
 *)

(** Send all request messages corresponding to the invocations on the
  requestQ.  Return the new bottomQ and requestQ, where the requestQ is now
  empty and the bottomQ has the appropriate bottom-half handlers tossed on 
*)
let rec send_available cout tree ~bottomQ ~requestQ =
  try
    let ((node,key),requestQ) = Fqueue.pop requestQ in
    let request = 
      if PTree.is_leaf node
      then ReconRqst_Full 
	{ rf_prefix = key;
	  rf_elements = PTree.elements tree node;
	} 
      else ReconRqst_Poly 
	{ rp_prefix = key;
	  rp_size = PTree.size node;
	  rp_samples = PTree.svalues node;
	}
    in
    marshal_noflush cout request;
    send_available cout tree
      ~bottomQ:(Fqueue.push (node,key) bottomQ) 
      ~requestQ
  with
      Fqueue.Empty -> 
	marshal_noflush cout Flush;
	incr flushcount;
	cout#flush; 
	(bottomQ,requestQ)

(** Read in all the response messages, matching them with the appropriate
   bottom-half handler on the bottomQ.  Return a new bottomQ and requestQ
   with the bottomQ empty and the appropriate requests added to the
   requestQ.   The returned bottomQ is always empty.
 *)
let rec read_available cin cout tree ~bottomQ ~requestQ set =
  try
    let ((node,key),bottomQ) = Fqueue.pop bottomQ  in
    let reply = unmarshal cin in
    begin 
      match reply.msg with
	| SyncFail ->
 	    if PTree.is_leaf node then
 	      raise (Bug ("Unexpected error.  Syncfail received" ^
 			  "at leaf node"));
 	    let children = PTree.child_keys tree key in
 	    let nodes = 
	      List.map 
 		~f:(fun key -> try PTree.get_node_key tree key
 		    with Not_found -> 
 		      raise (Bug ("Client.read: PTree.get_node_key " ^
 				  "should not fail")))
 		children in
 	    let requestQ = List.fold_left2 ~f:(fun q key node -> 
 						 Fqueue.push (node,key) q) 
 			     children nodes ~init:requestQ in 
 	    read_available cin cout tree ~bottomQ ~requestQ set

	| Elements elements -> 
	    read_available cin cout tree ~bottomQ ~requestQ 
 	    (Set.union set elements)
	    
	| FullElements elements ->
	    let local = PTree.get_zzp_elements tree node in
	    let localdiff = Set.diff local elements in
	    let remotediff = Set.diff elements local in
	    marshal_noflush cout (Elements localdiff);
	    read_available cin cout tree ~bottomQ ~requestQ
	      (Set.union set remotediff)


	| _ -> failwith ( "Unexpected message: " ^
 			  msg_to_string reply.msg )
    end

  with
      Fqueue.Empty ->
 	let (bottomQ,requestQ) = send_available cout tree 
				   ~bottomQ ~requestQ in
 	if Fqueue.is_empty bottomQ 
 	then ( 
	  marshal cout Done; 
	  set 
	) else  (
	  read_available cin cout tree ~bottomQ ~requestQ set
	)

(* Main reconciliation code *)
let reconcile tree cin cout =
  flushcount := 0; (* number of round-trips *)
  let startkey = Bitstring.create 0 in
  let requestQ = Fqueue.push (PTree.root tree, startkey) Fqueue.empty
  and bottomQ = Fqueue.empty in
  let rval = read_available cin cout tree ~bottomQ ~requestQ Set.empty in
  rval


let run_reconciliation tree ~partner ~self = 
  (* change this to depend on the address type *)
  let s = Unix.socket 
	    ~domain:Unix.PF_INET 
	    ~kind:Unix.SOCK_STREAM 
	    ~protocol:0 
  in
  let run () =
    Unix.connect s ~addr:partner;
    let cin = Channel.sys_in_from_fd s
    and cout = Channel.sys_out_from_fd s in

    signore (plerror 4 "Starting reconciliation");
    let remote_config = ReconCS.handle_config cin cout in

    let http_port = ReconCS.config_get_http_port remote_config in
    let remote_address = ReconCS.change_port partner http_port in

    let received_set = reconcile tree cin cout in
    (received_set,remote_address)
  in
  protect ~f:run ~finally:(fun () -> Unix.close s)

