(** Executable: server process that handles database and database queries. *)

open StdLabels
open MoreLabels
open Printf
open Common
open Packet
module Unix = UnixLabels
open Unix
open DbMessages

module Keydb = Keydb.Make(struct 
			    let withtxn = !Settings.transactions
			    and cache_bytes = !Settings.cache_bytes
			    and pagesize = !Settings.pagesize
			    and dbdir = !Settings.dbdir
			    and dumpdir = !Settings.dumpdir
			  end)


(* Simple server code for handling DB requests.  This is the main control
   code for the DB. *)

let withtxn = !Settings.transactions 
let dbdir = !Settings.dbdir
let () = 
  if not withtxn then 
    failwith "Running sks_db without transactions is no longer supported."


let websock = Eventloop.create_sock (ADDR_INET (inet_addr_any,http_port))
let () = 
  if Sys.file_exists db_command_name 
  then Unix.unlink db_command_name
let comsock = Eventloop.create_sock db_command_addr

let amp = Str.regexp "&"

let chsplit c s = 
  let eqpos = String.index s c in
  let first = Str.string_before s eqpos 
  and second = Str.string_after s (eqpos + 1) in
  (first, Wserver.decode second)

let eqsplit s = chsplit '=' s 

type request_kind = VIndex | Index | Get

type request = { kind: request_kind;
		 search: string list;
		 fingerprint: bool;
		 exact: bool;
	       }

let default_request = { kind = Index;
			search = [];
			fingerprint = false;
			exact = false;
		      }

let rec request_of_oplist ?(request=default_request) oplist = 
  match oplist with
      [] -> request
    | hd::tl -> 
	let new_request =
	  match hd with
	    | ("op","index") -> {request with kind = Index };
	    | ("op","vindex") -> {request with kind = VIndex };
	    | ("op","get") -> {request with kind = Get};
	    | ("search",s) ->  {request with search = 
				  Str.split Wserver.whitespace 
				    (String.lowercase s)
			       };
	    | ("fingerprint","on") ->  {request with fingerprint = true};
	    | ("fingerprint","off") ->  {request with fingerprint = false};
	    | ("exact","on") ->  {request with exact = true};
	    | ("exact","off") ->  {request with exact = false};
	    | _ -> request
	in
	request_of_oplist tl ~request:new_request


(****************************************************************************)
(** Database checkpointing and syncing *)

let sync () = 
  ignore (perror "Syncing database");
  Keydb.sync ();
  ignore (perror "Syncing complete")

let sync_interval = !Settings.db_sync_interval

let checkpoint () = 
  ignore (perror "Checkpointing database");
  Keydb.checkpoint ();
  ignore (perror "Checkpointing complete")
  
let checkpoint_interval = !Settings.checkpoint_interval


(***************************************************************)
(*  Helper functions for http request handler   ****************)
(***************************************************************)

(** sorts keys by time, dropping keys with no time *)
let tsort_keys keys = 
  let kpairs = 
    List.fold_left ~init:[] keys
      ~f:(fun list key -> 
	    try
	      let ki = ParsePGP.parse_pubkey_info (List.hd key) in
	      (ki.pk_ctime,key)::list
	    with
	      | Sys.Break as e -> raise e
	      | e -> list
	 )
  in
  let kpairs = List.sort ~cmp:compare kpairs in
  List.map ~f:snd kpairs

let html_quote string = 
  let sin = new Channel.string_in_channel string 0 in
  let sout = Channel.new_buffer_outc (String.length string + 10) in
  try
    while true do
      match sin#read_char with
	| '<' -> sout#write_string "&lt;"
	| '>' -> sout#write_string "&gt;"
	| '&' -> sout#write_string "&amp;"
	| '"' -> sout#write_string "&quot;"
	| c -> sout#write_char c  
    done;
    ""
  with
      Channel.TooShort _ ->
	sout#contents

(******************************************************************)
(******************************************************************)

let handle_request request =
  match request.kind with
    | Get -> 
	begin
	  try
	    let keyid_str = List.hd request.search in
	    let keyid = Fingerprint.keyid32_of_string keyid_str in
	    flush Pervasives.stdout;
	    let keys = Keydb.get_by_keyid keyid in
	    if keys = [] then failwith "No key found";
	    let keystr = Key.to_string_multiple keys in
	    let aakeys = Armor.encode_pubkey_string keystr in
	    HtmlTemplates.page  
	      ~title:(sprintf "Public Key Server -- Get ``%s ''" keyid_str)
	      ~body:(sprintf "\r\n<pre>\r\n%s\r\n</pre>\r\n" aakeys)
	  with
	    | Sys.Break as e -> raise e
	    | e -> 
		ignore (eperror e "Key not found");
		HtmlTemplates.page ~title:"Key Not Found" ~body:""
	end
    | Index -> 
	begin
	  try
	    let keys = Keydb.get_by_words ~max:!Settings.max_matches 
			 request.search in
	    if List.length keys = 0 then failwith "No keys retrieved";
	    let keys = tsort_keys keys in
	    let output = 
	      List.map keys 
		~f:(fun key -> 
		      try
			let meta = Fingerprint.from_key key in
			let keyid = Fingerprint.keyid_to_string ~short:true 
				      meta.Fingerprint.keyid in
			let link = HtmlTemplates.link 
				     ~hostname:!Settings.hostname
				     ~port:http_port ~keyid in
			let userids = List.map ~f:html_quote (Key.get_ids key) in
			let pki = ParsePGP.parse_pubkey_info (List.hd key) in
			let keystr = HtmlTemplates.keyinfo_pks pki ~keyid ~link 
				       ~userids in
			if request.fingerprint then
			  let fingerprint = 
			    HtmlTemplates.fingerprint 
			      ~fp:(Fingerprint.fp_to_string (meta.Fingerprint.fp))
			  in
			  Some (keystr,Some fingerprint)
			else
			  Some (keystr,None)
		      with
			| Sys.Break as e -> raise e
			| e ->
			    ignore (eplerror 2 e 
				      "Unable to print key from query '%s'"
				      (String.concat ~sep:" " request.search));
			    None
		   )
	    in
	    let output = filter_opts output in
	    let output = List.fold_left ~init:[] output 
			   ~f:(fun list (l1,l2) -> match l2 with
				   None -> l1::list
				 | Some l2 -> l1::l2::list)
	    in
	    let pre = HtmlTemplates.preformat_list 
			(HtmlTemplates.keyinfo_header :: output)
	    in
	    HtmlTemplates.page ~body:pre
	      ~title:(sprintf "Search results for '%s'" 
			(String.concat ~sep:" " request.search))

	  with
	    | Invalid_argument "Insufficiently specific words" ->
		HtmlTemplates.page ~body:"Provide at least one more specific keyword."
		  ~title:(sprintf "Insufficiently sufficient search: %s" 
			    (String.concat ~sep:" " request.search))

	    | Invalid_argument "Too many responses" ->
		HtmlTemplates.page ~body:"Too many matches.  Please restrict your query further."
		  ~title:(sprintf "Insufficiently sufficient search: %s" 
			    (String.concat ~sep:" " request.search))
		
	    | e -> 
		Eventloop.reraise e;
		ignore (eplerror 3 e "No keys found in Index query");
		HtmlTemplates.page ~body:""
		  ~title:(sprintf "No keys found: %s" 
			    (String.concat ~sep:" " request.search))
	end
    | VIndex -> 
	HtmlTemplates.page ~title:"Verbose Index not implemented" ~body:""

let string_to_oplist s = 
  let s = Wserver.strip s in 
  let (_,op_string) = chsplit '?' s in
  let oplist = Str.split amp op_string in
  List.map ~f:(chsplit '=') oplist

let bool_to_string b = if b then "true" else "false"
let print_request cout r = 
  fprintf cout "   kind: %s\n" (
    (function Index -> "index" | VIndex -> "vindex" | Get -> "get") r.kind);
  fprintf cout "   fingerprint: %s\n" (bool_to_string r.fingerprint);
  fprintf cout "   exact: %s\n" (bool_to_string r.exact);
  fprintf cout "   search: %s\n"
    (MList.to_string ~f:(fun x -> x) r.search)

let get_keystrings_from_hashes hashes = 
  let rec loop hashes keystrings = match hashes with
      [] -> keystrings
    | hash::tl -> 
	try 
	  let keystring = Keydb.get_keystring_by_hash hash in
	  loop tl (keystring::keystrings)
	with
	    e ->
	      ignore (eplerror 2 e "Error fetching key from hash %s" 
			(KeyHash.hexify hash));
	      loop tl keystrings
  in
  loop hashes []



(** Handler for HTTP requests *)
let webhandler addr msg cout = 
  match msg with 
    | Wserver.GET (request,headers) ->
	ignore (plerror 5 "Get request received");
	let request = request_of_oplist (string_to_oplist request) in
	cout#write_string (handle_request request);
	"text/html"
    | Wserver.POST (request,headers,body) ->
	ignore (plerror 5 "Post request received");
	let request = Wserver.strip request in
	match request with
	    "/pks/add" ->
	      ignore (plerror 5 "Handling /pks/add"); 
	      let keytext = Scanf.sscanf body "keytext=%s" (fun s -> s) in
	      let keytext = Wserver.decode keytext in
	      let keys = Armor.decode_pubkey keytext in
	      cout#write_string "<html><body>";
	      List.iter keys
		~f:(fun key -> 
		      try
			Keydb.add_key_merge ~newkey:true key;
			cout#write_string "Key Accepted<br>"
		      with
			| Bdb.Key_exists as e ->
			    cout#write_string 
			    ("Add failed: identical key already " ^
			     "exists in database<br>");
			    signore (eperror e "Key add failed")
			| e -> 
			    Eventloop.reraise e;
			    cout#write_string "Add failed<br>"; 
			    signore (eperror e "Key add failed")
		   );
	      cout#write_string "</html></body>";
	      "text/html"
	  | "/pks/hashquery" ->
	      ignore (plerror 5 "Handling /pks/hashquery"); 
	      let sin = new Channel.string_in_channel body 0 in
	      let hashes = 
		CMarshal.unmarshal_list ~f:CMarshal.unmarshal_string sin
	      in
	      let keystrings = get_keystrings_from_hashes hashes in
	      signore (perror "%d keys found" (List.length keystrings));
	      CMarshal.marshal_list ~f:CMarshal.marshal_string cout 
		keystrings;
	      "pgp/keys" (* This is a bogus content-type *)
	  | _ ->
	      cout#write_string (HtmlTemplates.page 
				   ~title:"Unexpected POST request" 
				   ~body:"");
	      "text/html"


(** Prepare handler for use with eventloop by transforming system
  channels to Channel objects and by returning empty list instead 
  of unit *)
let eventify_handler handle = 
  (fun addr cin cout ->
     let cin = (new Channel.sys_in_channel cin)
     and cout = (new Channel.sys_out_channel cout) in
     handle addr cin cout;
     []
  )

(** Handler for commands coming off of the db_command_addr *)
let command_handler addr cin cout = 
  match (unmarshal cin).msg with
    | LogQuery (count,timestamp) -> 
	let logresp = Keydb.logquery ~maxsize:count timestamp in
	let length = List.length logresp in
	if length > 0 then
	  ignore (plerror 3 "Sending LogResp size %d" length);
	marshal cout (LogResp logresp);

    | WordQuery words -> 
	ignore (plerror 3 "Handling WordQuery");
	let keys = Keydb.get_by_words ~max:!Settings.max_matches words in
	marshal cout (Keys keys)

    | Keys keys ->  
	marshal cout (Ack 0);
	(try Keydb.add_multi_keys_merge keys
	 with e -> ignore (eplerror 2 e "Key addition failed"))

    | DeleteKey hash ->
	ignore (plerror 3 "Handling DeleteKey");
	( try
	    let hash = RMisc.truncate hash KeyHash.hash_bytes in
	    let key = Keydb.get_by_hash hash in
	    Keydb.delete_key ~hash key;
	    marshal cout (Ack 0);
	  with
	      e -> 
		marshal cout (Ack (-1));
		raise e
	)
    | HashRequest hashes ->
	ignore (plerror 3 "Handling HashRequest");
	let keys = 
	  List.fold_left hashes ~init:[]
	    ~f:(fun list hash ->
		  try (Keydb.get_by_hash hash)::list
		  with 
		      Not_found -> 
			ignore (plerror 2 "Requested key %s not found"
				  (Utils.hexstring hash));
			list
	       )
	in
	ignore (plerror 3 "Returning set of %d keys" (List.length keys));
	marshal cout (Keys keys)


    | Config (s,cvar) ->
	ignore (plerror 4 "Received config message");
	(match (s,cvar) with
	   | ("checkpoint", `none) ->
	       checkpoint ()
	   | (str,value) ->
	       ignore (perror "Unexpected config request <%s>" str)
	)
	

    | m -> 
	marshal cout ProtocolError;
	ignore (perror "Unexpected (%s) message" (msg_to_string m))



(****************************************************************************)

(** Transmit enqueued keys to other hosts *)
let rec transmit_keys () = 
  (* Note: withtxn is ignored here.  Transactions are always used *)
  let dbs = Keydb.get_dbs () in
  let txn = Bdb.Txn.txn_begin dbs.Keydb.dbenv None [] in
  try
    match (try Some (Keydb.dequeue_key ~txn)
	   with Not_found -> None)
    with
      | Some (time,key) -> 
	  let body = Armor.encode_pubkey key in
	  let to_header = ("To", String.concat ~sep:", " 
			     (Membership.get_mailsync_partners ()))
	  in
	  let msg = { Sendmail.headers = 
			[ to_header;
			  "From", Settings.get_from_addr ();
			  "Reply-To", Settings.get_from_addr ();
			  "Errors-To", Settings.get_from_addr ();
			  "Subject","incremental";
			  "Precedence","list";
			  "Content-type", "application/pgp-keys";
			  "X-KeyServer-Sent",!Settings.hostname;
			] ;
		      Sendmail.body = body;
		    }
	  in
	  let string = Sendmail.msg_to_string msg in
	  signore (plerror 3 "Message transmitted for key %s"
		     (KeyHash.hexify (KeyHash.hash key)));
	  signore (plerror 6 "%s" string);
	  Sendmail.send msg;
	  Bdb.Txn.commit txn [];
	  signore (plerror 5 "transmission queue transaction committed");
	  transmit_keys ()
      | None -> 
	  Bdb.Txn.commit txn [];
	  []
  with
      e -> 
	Bdb.Txn.abort txn;
	raise e


(****************************************************************************)

let run () = 
  Keydb.open_dbs ();
  ignore (plerror 2 "Database opened");
  Eventloop.evloop

    (
      (if withtxn 
       then (Ehandlers.repeat_forever_simple checkpoint_interval checkpoint)
       else (Ehandlers.repeat_forever_simple sync_interval sync)) 
      @
      (Ehandlers.repeat_forever 10. 
	 (Eventloop.make_tc ~cb:transmit_keys ~timeout:0
	    ~name:"mail transmit keys" )
      )
      @
      (Ehandlers.repeat_forever 10. 
	 (Eventloop.make_tc ~name:"mailsync" ~timeout:0
	    ~cb:(Mailsync.load_mailed_keys 
		   ~addkey:(Keydb.add_key_merge ~newkey:false)))
      )
    )

    [(websock, Eventloop.make_th ~name:"webserver" 
	~timeout:!Settings.wserver_timeout
	~cb:(Wserver.accept_connection webhandler ~recover_timeout:1)
     );
     (comsock, Eventloop.make_th ~name:"command handler" 
	~timeout:!Settings.command_timeout
	~cb:(eventify_handler command_handler)
     );
     ]

let _ = 
  set_logfile ".db";
  ignore (plerror 1 "sks_db, SKS version %s" version); 
  ignore (plerror 1 "Copyright Yaron Minsky 2002"); 
  ignore (plerror 1 "Licensed under GPL.  See COPYING file for details"); 
  ignore (plerror 5 "http port: %d" http_port);
  protect ~f:run
    ~finally:(fun () -> 
		set_catch_break false;
		eprintf "Shutting down database\n"; 
		flush Pervasives.stderr;
		Keydb.close_dbs ()
	     )
