(** functions for constructing event handlers for use with [Eventloop] module *)
open StdLabels
open MoreLabels
open Printf

open Common
open Eventloop
module Unix = UnixLabels

(** Repeat callback ~request with a gap of redo_timeout, until
  either (test ()) is true or full_timeout has expired.
  In the former case, invoke success, int the latter, failure.
 
  Callbacks can return a list of events, which will be placed 
  on the queue upon their completion.  
 *)
let repeat_until ~redo_timeout ~full_timeout ~test
  ~init ~request ~success ~failure =
  init ();
  let start = Unix.gettimeofday () in
  let rec loop () = 
    let now = Unix.gettimeofday () in
    if test () 
    then success ()
    else if now > start +. full_timeout
    then failure ()
    else 
      (request ()) @ 
      [ Event (now +. redo_timeout, Callback loop) ]
  in
  let now = Unix.gettimeofday () in
  [ Event  (now, Callback loop) ]
  

(** returns smallest floating point number larger than the argument *)
let float_incr x = x +. x *. epsilon_float

let strftime time = 
  let tm = Unix.localtime time in
  sprintf "%02d:%02d:%02d" tm.Unix.tm_hour tm.Unix.tm_min tm.Unix.tm_sec
    

(** repeat provided callback forever, with one invocation occuring timeout
  seconds after the last one completed. *)
let repeat_forever timeout callback =
  let rec loop () = 
    let next_time = Unix.gettimeofday () +. timeout in
    [ Event (next_time, callback); 
      Event (float_incr next_time, Callback loop);
    ]
  in
  let now = Unix.gettimeofday () in
  [ Event (now, Callback loop); ]


let repeat_forever_simple timeout callback =
  repeat_forever timeout (Callback (fun () -> callback (); []))
