(** Executable: Builds up the key database from a multi-file database dump.
  This version works faster by virtue of not actually copying the keys out of
  the datbaase dump, and only storing the locations of those keys.
*)
open StdLabels
open MoreLabels
open Printf
open Arg
open Common
module Set = PSet.Set
module Unix = UnixLabels
open Packet

module Keydb = Keydb.Make(struct 
			    let withtxn = false
			    and cache_bytes = !Settings.cache_bytes
			    and pagesize = !Settings.pagesize
			    and dbdir = !Settings.dbdir
			    and dumpdir = !Settings.dumpdir
			  end)

let n = match !Settings.n with 0 -> 1 | x -> x
let maxkeys = n * 15000 
let dumpdir = !Settings.dumpdir

let lsdir dir = 
  let dirhandle = Unix.opendir dir in
  let rec loop accum = match (try Some (Unix.readdir dirhandle)
			 with End_of_file -> None)
  with
      Some fname -> loop (fname::accum)
    | None -> accum
  in
  loop []

let rec list_mapi list ~f = 
  let rec loop list i ~f = 
    match list with
	[] -> []
      | x::tl -> (f i x)::(loop tl (i + 1) ~f)
  in
  loop list 0 ~f

let fnames = List.filter (lsdir dumpdir)
	       ~f:(fun fname -> Filename.check_suffix fname ".pgp")
let fnames = List.map ~f:(fun fname -> Filename.concat dumpdir fname) fnames
let fnames = List.sort ~cmp:compare fnames
let nfnames = list_mapi ~f:(fun i fname -> (i,fname)) fnames

let timestr sec = 
  sprintf "%.2f min" (sec /. 60.)
    
let rec nsplit n list = match n with
    0 -> ([],list)
  | n -> match list with
	[] -> ([],[])
      | hd::tl -> 
	  let (beginning,ending) = nsplit (n-1) tl in
	  (hd::beginning,ending)

let rec batch_iter ~f n list = 
  match nsplit n list with
      ([],_) -> ()
    | (firstn,rest) -> f firstn; batch_iter ~f n rest

(******************************************************)

(** get single md using nextkey function *)
let get_keymd fnum nextkey = 
  match nextkey () with
      Some (pos,key) -> 
	let offset = { Keydb.fnum = fnum; 
		       Keydb.pos = pos; 
		     } 
	in
	Some (Keydb.key_to_metadata_large_offset offset key)
    | None -> 
	None

let rec get_keymds_rec ~max fnum nextkey accum =
  if max = 0 then (accum,0)
  else
    match get_keymd fnum nextkey with
	None -> (accum,max)
      | Some md -> 
	  get_keymds_rec ~max:(max-1) fnum nextkey 
	  (md::accum)

(** Fetches a collection of no more than max keys.  Returns (keys,bool), with
  the second argument being true of there is more to read from the given
  file. *)
let rec get_keymds ~max fnum nextkey = 
  get_keymds_rec ~max fnum nextkey []


(*
let get_keymds_inchan ~max fnum inchan =
  let cin = new Channel.sys_in_channel inchan in
  let nextkey = Key.pos_next_of_channel cin in
  get_keymds ~max fnum nextkey 
*)

let inchan_to_nextkey inchan = 
  let cin = new Channel.sys_in_channel inchan in
  Key.pos_next_of_channel cin

let rec get_keymds_list ~max nflist partial =
  match nflist with 
      [] -> (partial,[])
    | (fnum,nextkey)::tl -> 
	if max = 0 then (partial,nflist)
	else 
	  let (mds,remaining) = get_keymds ~max fnum nextkey in
	  (* printf "remaining = %d,nflist len = %d\n"
	    remaining (List.length nflist);  *)
	  flush stdout;
	  if remaining > 0 then (
	    (* file must be done with, so don't pass it on *)
	    get_keymds_list ~max:remaining tl (List.rev_append mds partial)
	  ) else (
	    (* file is not (necessarily) done, 
	       but we've got the key mds we need *)
	    (List.rev_append mds partial,nflist)
	  )

let get_keymds_list ~max nflist = get_keymds_list ~max nflist []

let dbtimer = MTimer.create ()
let timer = MTimer.create ()
let () = 
  if Sys.file_exists !Settings.dbdir then
    ignore (Unix.system (sprintf "rm -r %s" !Settings.dbdir));
  Unix.mkdir !Settings.dbdir 0o700;

  Keydb.open_dbs ();
  let filearray = Keydb.get_dump_filearray () in
  let nfarray = Array.mapi ~f:(fun i x -> (i,inchan_to_nextkey x))
		  filearray in
  let nflist = Array.to_list nfarray in

  printf "Loading %d keys at a time\n" maxkeys;

  protect 
    ~f:(fun () ->
	  let rec loop nflist = match nflist with 
	      [] -> ()
	    | nflist -> 
		MTimer.start timer;

		printf "Loading metadata...\n"; flush stdout;
		let (mds,nflist) = get_keymds_list ~max:maxkeys nflist in
		printf "   %d keys loaded, %d files left\n" 
		  (List.length mds) (List.length nflist);
		flush stdout;
		MTimer.start dbtimer; 
		Keydb.add_mds mds;
		MTimer.stop dbtimer;

		MTimer.stop timer;
		printf "   DB time:  %s.  Total time: %s.\n" 
		  (timestr (MTimer.read dbtimer)) 
		  (timestr (MTimer.read timer)); 
		flush stdout;
		loop nflist
	    
	  in
	  loop nflist
       )
    ~finally:(fun () -> Keydb.close_dbs)

