(** Logic for merging PGP keys with the same public key *)
open StdLabels
open MoreLabels
open Printf
open Common
open Packet

module Set = PSet.Set
module Map = PMap.Map

(*
  module PacketCompare = 
  struct 
  type t = packet 
  let compare = (fun x y -> compare 
  (x.packet_type,x.packet_body)
  (y.packet_type,y.packet_body)
  )
  end

  module Set = Set.Make(PacketCompare)
  module Map = Map.Make(PacketCompare)
*)				 


(** This is my understanding of the grammar of allowable public keys:

{[   ATOMS = v3_pubkey v4_pubkey signature pubsubkey uid e]}

   The above correspond to packet types, except for e, which corresponds to
   the empty string.

   Here's the grammar:

{[      KEY := V3 | V4
      V3 := v3_pubkey SIGLIST UIDLIST
      V4 := v4_pubkey SIGLIST UIDLIST SUBKEYLIST
      SIGLIST := e | signature SIGLIST
      UIDLIST := e | UID UIDLIST
      UID := uid SIGLIST | uid
      SUBKEYLIST := e | SUBKEY SUBKEYLIST
      SUBKEY := subkey SIGLIST ]}

(shouldn't the last one be:
  {[SUBKEY := subkey signature SIGLIST]}
 since there must be at least one signature?)

   My only purpose in doing this parsing is to allow for the proper merging
   of two public keys.  
   To merge two keys, I join the SIGLISTs and UIDLISTs and SUBKEYLISTs.
     + Merging SIGLISTs is straightforward: just concatentate the lists and drop
       duplicates.
     + Merging UIDLISTs and SUBKEYLISTs is somewhat more complicated.  I
       join siglists corresponding to the same UID.  

   The current implementation explicitly distinguishes between v3 and v4
   keys, which really it doesn't need to do as it presently stands.  But if a
   fuller handler of revocation becomes necessary, then distinguishing
   between the two may be necessary.

   There is no special handling of revocations --- I don't check if they're
   valid, and multiple revocations can pop up.
*)


(*******************************************************************)
(* Types for representing the structure of a key *)

type sigpair = packet * packet list

type v3key = { v3_key : packet;
     	       v3_revocations: packet list;
     	       v3_uids: sigpair list }

type v4key = { v4_key: packet;
     	       v4_selfsigs: packet list; (* may include revocations *)
     	       v4_uids: sigpair list;
     	       v4_subkeys: sigpair list;
     	     }

(** parsed key *)
type pkey = V3 of v3key | V4 of v4key (* a parsed key *)



let packets_equal p1 p2 = p1 = p2

(*
let packets_equal p1 p2 = 
  p1.packet_body = p2.packet_body && p1.packet_type = p2.packet_type

let ( <>@ ) p1 p2 = not (packets_equal p1 p2)
let ( =@ ) p1 p2 = packets_equal p1 p2
*)

(*******************************************************************)
(** Code for flattening out the above structure back to the original key *)

let rec flatten_sigpair_list list = match list with
    [] -> []
  |  (pack,sigs)::tl -> pack :: (sigs @ flatten_sigpair_list tl)

let flatten key = match key with
    V4 v4 ->
      v4.v4_key :: List.concat [ v4.v4_selfsigs; 
				 flatten_sigpair_list v4.v4_uids; 
				 flatten_sigpair_list v4.v4_subkeys ]
  | V3 v3 ->
      v3.v3_key :: List.concat [ v3.v3_revocations;
				 flatten_sigpair_list v3.v3_uids;
			       ]

(************************************************************)

let print_pkey key = match key with
    V4 k -> 
      printf "V4 Key: %d selfsigs, %d uids, %d subkeys\n" 
	(List.length k.v4_selfsigs)
	(List.length k.v4_uids)
	(List.length k.v4_subkeys)
  | V3 k ->
      printf "V3 Key: %d revocations, %d uids\n" 
	(List.length k.v3_revocations)
	(List.length k.v3_uids)


(*******************************************************************)

let get_version packet =  
  match packet.packet_type with 
      Public_Key_Packet -> int_of_char packet.packet_body.[0]
    | Signature_Packet -> int_of_char packet.packet_body.[0]
    | _ -> raise Not_found

let key_to_stream key = 
  let ptype_list = List.map ~f:(fun pack -> (pack.packet_type,pack)) key in
  Stream.of_list ptype_list




(*******************************************************************)
(*** Key Parsing ***************************************************)
(*******************************************************************)

let rec parse_keystr = parser
  | [< '(Public_Key_Packet,p) ; s >] ->
      match get_version p with
	| 4 ->
	    (match s with parser [< selfsigs = siglist; 
				    uids = uidlist; 
				    subkeys = subkeylist >]
		 -> V4 { v4_key = p; 
			 v4_selfsigs = selfsigs; 
			 v4_uids = uids; 
			 v4_subkeys = subkeys;
		       })
	| 2 | 3 -> 
	    (match s with parser [< revocations = siglist; 
				    uids = uidlist >] -> 
	       V3 { v3_key = p ; 
		    v3_revocations = revocations; 
		    v3_uids = uids;
		  })
	| _ -> failwith "Unexpected key packet version number"
and siglist = parser						                
  | [< '(Signature_Packet,p); tl = siglist >] -> p::tl
  | [< >] -> []
and uidlist = parser						                
  | [< '(User_ID_Packet,p); sigs = siglist; s >] -> 
      (p,sigs)::(match s with parser
		   | [< '(User_ID_Packet,p); sigs = siglist; tl = uidlist >] -> 
		       (p,sigs)::tl
		   | [< >] -> [])
  | [< >] -> []
and subkeylist = parser
  | [< '(Public_Subkey_Packet,p); sigs = siglist; tl = subkeylist >] -> 
      (p,sigs)::tl
  | [< >] -> []

(*******************************************************************)

let apply_opt ~f opt = match opt with
    None -> None
  | Some x -> Some (f x)


(*******************************************************************)
(*** Key Merging Code  *********************************************)
(*******************************************************************)

let set_of_list list = List.fold_left ~init:Set.empty list
			 ~f:(fun set x -> Set.add x set)

let dedup list = Set.elements (set_of_list list)

let merge_sigpairs pairs =
  let map = 
    List.fold_left pairs
      ~f:(fun map (pack,sigs) ->
	    try 
	      let old_sigs = Map.find pack map in
	      (* If front packet is already there, add in new sigs,
		 discarding duplicates *)
	      Map.add ~key:pack ~data:(dedup (old_sigs @ sigs)) map
	    with
		(* otherwise, add in data by itself *)
		Not_found -> Map.add ~key:pack ~data:sigs map)
      ~init:Map.empty
  in
  Map.fold ~f:(fun ~key:pack ~data:sigs list -> (pack,sigs)::list) map ~init:[] 

let merge_sigpair_lists l1 l2 = 
  merge_sigpairs (l1 @ l2)

(*******************************************************************)

let merge_pkeys key1 key2 = match (key1,key2) with
    (V3 _, V4 _) | (V4 _, V3 _) -> None
  | (V3 k1, V3 k2) ->
      if not (packets_equal k1.v3_key k2.v3_key)
      then None (* merge can only work if keys are the same *)
      else
	Some (V3 { v3_key = k1.v3_key;
		   v3_revocations = 
		     dedup (k1.v3_revocations @ k2.v3_revocations); 
		   v3_uids = merge_sigpair_lists k1.v3_uids k2.v3_uids;
		 })
  | (V4 k1, V4 k2) ->
      if not (packets_equal k1.v4_key k2.v4_key)
      then None (* merge can only work if keys are the same *)
      else
	Some (V4 { v4_key = k1.v4_key;
		   v4_selfsigs = dedup (k1.v4_selfsigs @ k2.v4_selfsigs); 
		   (* this might be wrong.  Must the revocations 
		      be separated out to go before the other self
		      signatures? *)
		   v4_uids = merge_sigpair_lists k1.v4_uids k2.v4_uids;
		   v4_subkeys = merge_sigpair_lists k1.v4_subkeys k2.v4_subkeys;
		 })

(*******************************************************************)
(*******************************************************************)
(*******************************************************************)

(** Checks whether any of self-sigs of parsed key have no-modify
  flag set *)
let pkey_has_nomodify pkey = match pkey with
    V3 _ ->  false
  | V4 v4key ->
      let rec testloop sigs = match sigs with 
	  [] -> false
	| hd::tl -> 
	    if ParsePGP.signature_has_nomodify hd then true
	    else testloop tl
      in
      let selfsigs = v4key.v4_selfsigs in
      testloop selfsigs

let key_to_pkey key = parse_keystr (key_to_stream key)

let merge key1 key2 = 
  try
    let pkey1 = key_to_pkey key1
    and pkey2 = key_to_pkey key2 in
    let mkey = merge_pkeys pkey1 pkey2 in
    apply_opt ~f:flatten mkey
  with 
      Stream.Failure | Stream.Error _ -> None

