(** Interface for dealing with underlying key database *)
open StdLabels
open MoreLabels
open Printf
open Common
module Set = PSet.Set

open Bdb
open Packet

module type DBSettings = 
sig
  val withtxn: bool
  val cache_bytes: int option
  val pagesize: int option
  val dbdir: string
  val dumpdir: string
end

module Make(Dbs: DBSettings) =
struct

  open Dbs

  let word_db_name = "word"
  let key_db_name = "key"
  let keyid_db_name = "keyid"
  let time_db_name = "time"
  let tqueue_db_name = "tqueue"

  let max_internal_matches = !Settings.max_internal_matches

  (**********************************************************)
  (*  Types  ************************************************)
  (**********************************************************)

  type 'a offset = { fnum: int; pos: 'a; }
      
  (** Stored key.  Can have a number of formats.  
    Eventually this may include death certificates 
  *)
  type skey = 
    | KeyString of string  
    | Key of packet list 
    | Offset of int offset
    | LargeOffset of int64 offset

  type dbdump =
      { directory: string;
	filearray: in_channel array; 
      }

  type dbstate = 
      { dbenv: Dbenv.t;
	key: Db.t;
	word: Db.t;
	keyid: Db.t; 
	time: Db.t;  
	tqueue: Db.t; (* queue of hashes that need 
			 to be transmitted to other hosts *)
	dump: dbdump; (* info @ dump files where initial 
			 keydump is stored *)
      }
  let dbstate = ref None
  exception No_db

  (***********************************************************************)

  let get_dbs () =
    match !dbstate with 
	None -> raise No_db
      | Some dbs -> dbs

  let get_dump_filearray () = 
    let dbs = get_dbs () in
    dbs.dump.filearray


  (***********************************************************************)
  (*  Key conversions ****************************************************)
  (***********************************************************************)

  let marshal_offset cout offset = 
    cout#write_int offset.fnum;
    cout#write_int offset.pos

  let unmarshal_offset cin = 
    let fnum = cin#read_int in
    let offset = cin#read_int in
    { fnum = fnum; pos = offset; }

  (***********************************************************************)

  let marshal_large_offset cout offset = 
    cout#write_int offset.fnum;
    cout#write_int64 offset.pos

  let unmarshal_large_offset cin = 
    let fnum = cin#read_int in
    let offset = cin#read_int64 in
    { fnum = fnum; pos = offset; }


  (***********************************************************************)

  let skey_of_string s = 
    let cin = new Channel.string_in_channel s 0 in
    match cin#read_byte with
	0 -> KeyString cin#read_rest
      | 1 -> Offset (unmarshal_offset cin)
      | 2 -> LargeOffset (unmarshal_large_offset cin)
      | _ -> failwith "Unexpected skey type"

  let skey_to_string skey = 
    let cout = Channel.new_buffer_outc 0 in
    (match skey with
	 KeyString s -> cout#write_byte 0; cout#write_string s
       | Key key -> cout#write_byte 0; Key.write key cout
       | Offset offset -> cout#write_byte 1; marshal_offset cout offset
       | LargeOffset offset -> cout#write_byte 2; 
	   marshal_large_offset cout offset
    );
    cout#contents

  let keystring_of_skey skey = match skey with
      KeyString s -> s
    | Key key -> Key.to_string key
    | Offset offset -> 
	let dbs = get_dbs () in
	if Array.length dbs.dump.filearray  = 0
	then failwith ("Key could not be fetched from offset: " ^
		       "No key dump found");
	if offset.fnum > Array.length dbs.dump.filearray 
	then failwith ("Key could not be fetched from offset: " ^
		       "File number exceeds number of dump files");
	let file = dbs.dump.filearray.(offset.fnum) in
	seek_in file offset.pos;
	let key = Key.get_of_channel (new Channel.sys_in_channel file) () in
	Key.to_string key
    | LargeOffset offset -> 
	let dbs = get_dbs () in
	if Array.length dbs.dump.filearray  = 0
	then failwith ("Key could not be fetched from offset: " ^
		       "No key dump found");
	if offset.fnum > Array.length dbs.dump.filearray 
	then failwith ("Key could not be fetched from offset: " ^
		       "File number exceeds number of dump files");
	let file = dbs.dump.filearray.(offset.fnum) in
	LargeFile.seek_in file offset.pos;
	let key = Key.get_of_channel (new Channel.sys_in_channel file) () in
	Key.to_string key

  let keystring_of_string string = 
    keystring_of_skey (skey_of_string string)

  let key_of_skey skey = 
    match skey with
	KeyString s -> Key.of_string s
      | Key key -> key
      | Offset offset -> 
	  let dbs = get_dbs () in
	  if Array.length dbs.dump.filearray  = 0
	  then failwith ("Key could not be fetched from offset: " ^
			 "No key dump found");
	  if offset.fnum > Array.length dbs.dump.filearray 
	  then failwith ("Key could not be fetched from offset: " ^
			 "File number exceeds number of dump files");
	  let file = dbs.dump.filearray.(offset.fnum) in
	  seek_in file offset.pos;
	  Key.get_of_channel (new Channel.sys_in_channel file) ()
      | LargeOffset offset -> 
	  let dbs = get_dbs () in
	  if Array.length dbs.dump.filearray  = 0
	  then failwith ("Key could not be fetched from offset: " ^
			 "No key dump found");
	  if offset.fnum > Array.length dbs.dump.filearray 
	  then failwith ("Key could not be fetched from offset: " ^
			 "File number exceeds number of dump files");
	  let file = dbs.dump.filearray.(offset.fnum) in
	  LargeFile.seek_in file offset.pos;
	  Key.get_of_channel (new Channel.sys_in_channel file) ()
	    

  let key_to_string key = skey_to_string (Key key)
  let key_of_string s = key_of_skey (skey_of_string s)

  (***********************************************************************)

  (** returns a list of all elements of the specified directory 
    with the given suffix *)
  let read_dir_suff dir suff = 
    let dh = Unix.opendir dir in
    let dirs = ref [] in
    while 
      match (try Some (Unix.readdir dh)
	     with End_of_file -> None)
      with
	  Some fname -> 
	    if Filename.check_suffix fname suff 
	    then dirs := fname::!dirs;
	    true
	| None -> 
	    false
    do () done;
    List.rev !dirs 

  (***********************************************************************)

  (** Initialization code for database *)
  let open_dbs () = 
    ignore (plerror 3 "Opening KeyDB database");
    match !dbstate with
	Some x -> failwith ("Keydb.open_dbs: Attempt to open when " ^
			    "close_dbs hasn't been called")
      | None ->
	  let dbenv =  Dbenv.create [] in
	  ( match cache_bytes with None -> ()
	      | Some cache_bytes -> Dbenv.set_cachesize dbenv 
		  ~gbytes:0 ~bytes:cache_bytes ~ncache:0);
	  Dbenv.dopen dbenv dbdir 
	    ( [ Dbenv.INIT_MPOOL; Dbenv.CREATE;]
	      @ ( if withtxn then [ Dbenv.INIT_TXN; Dbenv.RECOVER ]
		  else [] ) )
	    0o600;

	  let key = Db.create ~dbenv [] in
	  (match pagesize with None -> ()
	     | Some pagesize -> Db.set_pagesize key pagesize);
	  Db.dopen key key_db_name Db.BTREE [Db.CREATE] 0o600;

	  let word = Db.sopen ~dbenv word_db_name Db.BTREE
		       ~moreflags:[Db.DUPSORT] [Db.CREATE] 0o600 in
	  let keyid =  Db.sopen ~dbenv keyid_db_name Db.BTREE
			 ~moreflags:[Db.DUPSORT] [Db.CREATE] 0o600 in
	  let time = Db.sopen ~dbenv time_db_name Db.BTREE 
		       ~moreflags:[Db.DUPSORT] [Db.CREATE] 0o600 in
	  let tqueue = Db.sopen ~dbenv tqueue_db_name Db.BTREE
			 ~moreflags:[] [Db.CREATE] 0o600 in

	  (** Sets up array of dump files for entries where 
	    file offset is stored instead of key contents *)
	  let dump = 
	    let dir = dumpdir in
	    if (Sys.file_exists dir &&
		(Unix.stat dir).Unix.st_kind = Unix.S_DIR)
	    then
	      let pgpfiles = read_dir_suff dir ".pgp" in
	      let pgpfiles = List.sort ~cmp:compare pgpfiles in
	      let pgpfiles = 
		List.map ~f:(fun f -> Filename.concat dir f) pgpfiles in
	      let pgpfiles = Array.of_list pgpfiles in
	      { directory = dir;
		filearray = 
		  Array.map 
		    ~f:(open_in_gen [Open_rdonly; Open_binary] 0o600) 
		    pgpfiles
	      }
	    else
	      { directory = "";
		filearray = Array.make 0 stdin;
	      }
	  in

	  if withtxn then Txn.checkpoint dbenv ~kbyte:0 ~min:0 [];
	  dbstate := Some { dbenv = dbenv;
			    word = word;
			    key = key; 
			    keyid = keyid;
			    time = time;
			    dump = dump;
			    tqueue = tqueue;
			  }

  (** create dump info from name of directory containing dump *)
  let open_dump dumpdir = 
    let pgpfiles = read_dir_suff dumpdir ".pgp" in
    let pgpfiles = List.sort ~cmp:compare pgpfiles in
    let pgpfiles = Array.of_list pgpfiles in
    let dump = { directory = dumpdir;
		 filearray = Array.map ~f:open_in_bin pgpfiles
	       }
    in
    let dbs = get_dbs () in
    if Array.length dbs.dump.filearray <> 0 
    then failwith "Attempt to reopen dumpdir"
    else dbstate := Some { dbs with dump = dump }
      

  (***********************************************************************)

  let close_dbs () = match !dbstate with
      None -> raise No_db
    | Some dbs ->
	Db.close dbs.key;
	Db.close dbs.word;
	Db.close dbs.time;
	Db.close dbs.keyid;
	Db.close dbs.tqueue;
	Dbenv.close dbs.dbenv;
	dbstate := None

  (***********************************************************************)

  let sync () = 
    let dbs = get_dbs () in
    Db.sync dbs.key;
    Db.sync dbs.word;
    Db.sync dbs.time;
    Db.sync dbs.keyid;
    Db.sync dbs.tqueue

  (***********************************************************************)

  let checkpoint () = 
    if withtxn then 
      let dbs = get_dbs () in
      Txn.checkpoint dbs.dbenv ~kbyte:(1024 * 10) ~min:0 []

  (***********************************************************************)

  let is_alnum char = 
    let num = int_of_char char in
    (num >= int_of_char 'A' && num <= int_of_char 'Z') ||
    (num >= int_of_char 'a' && num <= int_of_char 'z') ||
    (num >= int_of_char '0' && num <= int_of_char '9') ||
    (num >= 192 && num <= 255)

  (***********************************************************************)
  (** Entry preparation code: utilities for formatting data for placement in
    database *)
  (***********************************************************************)

  let rec extract_words_rec s ~start ~len partial = 
    if start + len = String.length s 
    then ( if len = 0 then partial 
	   else Set.add (String.sub s start len) partial )
    else (
      if is_alnum s.[start + len]
      then extract_words_rec s ~start ~len:(len + 1) partial
      else ( if len = 0 
	     then extract_words_rec s ~start:(start + 1) ~len partial
	     else extract_words_rec s ~start:(start + len)  ~len:0
	       (Set.add (String.lowercase (String.sub s start len)) partial)
	   )
    )

  (**  returns the set of words found in string s *)
  let extract_word_set s = 
    extract_words_rec s ~start:0 ~len:0 Set.empty

  (** returns a list of words found in string s *)
  let extract_words s =
    Set.elements (extract_word_set s)
      
  (** returns the list of words found in all of the userid strings of given
    key*)
  let key_extract_words key = 
    let userids = Key.get_ids key in
    let wordsets = List.map ~f:extract_word_set userids in
    Set.elements (List.fold_left ~init:Set.empty ~f:Set.union
		    wordsets)

  (***********************************************************************)

  let float_to_string f = 
    let cout = Channel.new_buffer_outc 8 in
    cout#write_float f;
    cout#contents

  let float_of_string s = 
    let cin = new Channel.string_in_channel s 0 in
    cin#read_float

  let event_to_string event = 
    let cout = Channel.new_buffer_outc 9 in
    ( match event with
	  Add hash -> cout#write_byte 0; cout#write_string hash
	| Delete hash -> cout#write_byte 1; cout#write_string hash
    );
    cout#contents 

  let event_of_string string = 
    let cin = new Channel.string_in_channel string 0 in
    match cin#read_byte with
	0 -> Add cin#read_rest
      | 1 -> Delete cin#read_rest
      | _ -> failwith "Failure parsing event string"

  let flatten_array_of_lists a = 
    Array.of_list (List.flatten (Array.to_list a))

  let flatten_array_of_lists a = 
    let total_length = 
      Array.fold_left ~init:0
	~f:(fun sum list -> sum + List.length list) a
    in
    let newarray = Array.make total_length ("","") in
    let ctr = ref 0 in
    Array.iter a 
      ~f:(List.iter ~f:(fun el -> newarray.(!ctr) <- el; incr ctr));
    newarray

      
  (***********************************************************************)

  let jcursor_get_all ?max c =
    let rec loop max list = 
      if max = 0 then list
      else (
	match (try Some (Cursor.get c Cursor.NULL []) 
	       with Not_found -> None)
	with 
	    Some (key,data) -> loop (max - 1) (data :: list)
	  | None -> list
      )
    in
    loop 
      (match max with Some max -> max | None -> max_int)
      []


  (***********************************************************************)
  (*  Access methods  ***************************************************)
  (***********************************************************************)

  (** retrieve keys based on words found in uid strings *)
  let get_by_words ~max wordlist = 
    let dbs = get_dbs () in
    try
      let cursors = List.map ~f:(fun word -> 
				   let c = Cursor.create dbs.word in
				   ignore (Cursor.init c word []);
				   c )
		      wordlist in
      let lengths = List.map ~f:Cursor.count cursors in
      if MList.min lengths > max_internal_matches
      then raise (Invalid_argument "Insufficiently specific words");
      let cj = Cursor.join dbs.key cursors [] in
      let keystrings = jcursor_get_all cj in
      if List.length keystrings > max then
	raise (Invalid_argument "Too many responses")
      else
	List.map ~f:key_of_string keystrings 
    with
	Not_found -> []


  (***********************************************************************)

  (** retrieve hashes of keys that were inserted at or before a specified
    time *)
  let get_by_time ~max time = 
    let dbs = get_dbs () in
    let c = Cursor.create dbs.time in
    let rec loop list =
      match (try Some (Cursor.get c Cursor.NEXT []) 
	     with Not_found -> None)
      with 
	| Some (key,data) -> loop (data::list)
	| None -> List.rev list
    in
    try
      let timestr = float_to_string time in
      let (key,data) = Cursor.init_range c timestr [] in
      loop [data]
    with
	Not_found -> []

  (***********************************************************************)

  (** returns true iff db contains specified hash *)
  let has_hash hash =
    let dbs = get_dbs () in
    try
      let data = Db.get dbs.key hash [] in
      true
    with
	Not_found -> false

  (***********************************************************************)

  let get_skeystring_by_hash hash = 
    let dbs = get_dbs () in
    Db.get dbs.key hash []

  let get_keystring_by_hash hash = 
    keystring_of_string (get_skeystring_by_hash hash)

  (***********************************************************************)

  (** retrieves key by hash *)
  let get_by_hash hash = 
    key_of_string (get_skeystring_by_hash hash)


  (** Verification functions *)

  let check_word_hash_pair ~word ~hash = 
    let dbs = get_dbs () in
    let c = Cursor.create dbs.word in
    try 
      Cursor.init_both c ~key:word ~data:hash [];
      true
    with
	Not_found -> false

  let check_keyid_hash_pair ~keyid ~hash = 
    let dbs = get_dbs () in
    let c = Cursor.create dbs.keyid in
    try 
      Cursor.init_both c ~key:keyid ~data:hash [];
      true
    with
	Not_found -> false
    
  (***********************************************************************)

  (** removes key from database, also removing key from all indices *)
  let delete_key ?txn ?hash key = 
    let dbs = get_dbs () in
    let hash = match hash with 
	Some hash -> hash
      | None -> KeyHash.hash key 
    in
    let words = key_extract_words key in

    (* delete from worddb *)
    let wordc = Cursor.create ?txn dbs.word in
    let del_word word = 
      Cursor.init_both wordc ~key:word ~data:hash [];
      Cursor.del wordc;
    in
    protect ~f:(fun () -> List.iter ~f:(fun x -> del_word x) words)
      ~finally:(fun () -> Cursor.close wordc);
    
    (* delete from keyid *)
    let keyid = Fingerprint.keyid_from_key ~short:true key in
    let keyidc = Cursor.create ?txn dbs.keyid in
    protect ~f:(fun () -> 
		  Cursor.init_both keyidc ~key:keyid ~data:hash [];
		  Cursor.del keyidc)
      ~finally:(fun () -> Cursor.close keyidc);

    (* delete from key *)

    let keyc = Cursor.create ?txn dbs.key in
    protect ~f:(fun () -> 
		  ignore (Cursor.init keyc hash []);
		  Cursor.del keyc)
      ~finally:(fun () -> Cursor.close keyc);


    let timestr = float_to_string (Unix.gettimeofday ()) 
    and eventstr = event_to_string (Delete hash) in
    Db.put ?txn dbs.time ~key:timestr ~data:eventstr [Db.NODUPDATA]


  (***********************************************************************)
      
  let get_keystrings_by_hashes hashes = 
    (* sort to improve performance, although this should 
       only really help for very large lists. *)
    let hashes = List.sort ~cmp:compare hashes in 
    let keystr_opts = 
      List.map ~f:(fun hash -> 
		     try Some (get_keystring_by_hash hash) 
		     with Not_found -> None)
	hashes 
    in
    MList.strip_opt keystr_opts

  (***********************************************************************)

  let iter ~f =
    let dbs = get_dbs () in
    let c = Cursor.create dbs.key in
    let rec loop get_type =
      match (try Some (Cursor.get c get_type []) with Not_found -> None)
      with
	| Some (key,data) -> f ~key ~data; loop Cursor.NEXT 
	| None -> ()
    in
    loop Cursor.FIRST
      

  (***********************************************************************)

  let keyiter ~f =
    let dbs = get_dbs () in
    let c = Cursor.create dbs.key in
    let rec loop get_type =
      match (try Some (Cursor.get_keyonly c get_type []) 
	     with Not_found -> None)
      with
	| Some key -> f key; loop Cursor.NEXT 
	| None -> ()
    in
    loop Cursor.FIRST

  (***********************************************************************)

  let get_hashes_by_keyid keyid = 
    let dbs = get_dbs () in 
    let c = Cursor.create dbs.keyid in
    let rec loop list =
      match (try Some (Cursor.get c Cursor.NEXT_DUP [])
	     with Not_found -> None)
      with 
	| Some (key,data) -> loop (data::list)
	| None -> List.rev list
    in
    try
      let first = Cursor.init c keyid [] in
      let hashes = loop [first] in
      hashes
    with
	Not_found -> []

  let get_skeystrings_by_keyid keyid = 
    let dbs = get_dbs () in
    let hashes = get_hashes_by_keyid keyid in
    MList.strip_opt 
      (List.map ~f:(fun hash -> 
		     try Some (get_skeystring_by_hash hash)
		     with Not_found -> 
		       signore (perror "%s: %s Not_found"
				  "Failed conversion of hash to skeystring"
				  (KeyHash.hexify hash));
		       None
		   )
	 hashes)

  let get_by_keyid keyid = 
    let skeystrings = get_skeystrings_by_keyid keyid in
    List.map ~f:key_of_string skeystrings


  let logquery ?(maxsize=5000) timestamp = 
    
    let dbs = get_dbs () in
    let c = Cursor.create dbs.time in
    try
      let (timestr,eventstr) = 
	Cursor.init_range c (float_to_string timestamp) [] in
      let fst_time = float_of_string timestr in
      let fst_event = event_of_string eventstr in
      assert (fst_time >= timestamp);
      let rec loop count list = match count with
	  0 -> List.rev list
	| _ -> 
	    try
	      let (time,event) = Cursor.get c Cursor.NEXT [] in
	      let (time,event) = (float_of_string time, 
				  event_of_string event) 
	      in loop (count - 1) ((time,event)::list)
	    with
		Not_found -> List.rev list
      in
      if fst_time = timestamp then loop maxsize [] 
      else loop (maxsize - 1) [(fst_time,fst_event)]
    with 
	Not_found -> []


  (***********************************************************************)

  let create_hashstream () = 
    let dbs = get_dbs () in
    let c = Cursor.create dbs.key in
    let first = Cursor.get_keyonly c Cursor.FIRST [] in
    let close () = Cursor.close c in
    let next () = (try Some (Cursor.get_keyonly c Cursor.NEXT [])
		   with Not_found -> None) in
    let stream = SStream.make ~first next in
    (stream,close)

  let create_hash_skey_stream () = 
    let dbs = get_dbs () in
    let c = Cursor.create dbs.key in
    let first = Cursor.get c Cursor.FIRST [] in
    let close () = Cursor.close c in
    let next () = (try Some (Cursor.get c Cursor.NEXT [])
		   with Not_found -> None) in
    let stream = SStream.make ~first next in
    (stream,close)


  (***********************************************************************)

  let last_ts () = 
    let dbs = get_dbs () in
    let c = Cursor.create dbs.time in
    protect ~f:(fun () -> float_of_string (Cursor.get_keyonly c 
					     Cursor.LAST []))
      ~finally:(fun () -> Cursor.close c)


  (**************************************************************)
  (**  Functions for updating key database *)
  (**************************************************************)

  (**********************************************************)

  (** Add key to transmission queue for sending to other 
    (non-SKS) keyservers. *)
  let enqueue_key ~txn key =
    let dbs = get_dbs () in
    let c = Cursor.create ~txn dbs.tqueue in
    let run () = 
      let timestr = float_to_string (Unix.gettimeofday ()) in
      Cursor.kput c ~key:timestr ~data:(key_to_string key) Cursor.KEYLAST
    in
    protect ~f:run ~finally:(fun () -> Cursor.close c)

  (** Extract key from transmission queue for receiving from
    (non-SKS) keyservers. *)
  let dequeue_key ~txn =
    let dbs = get_dbs () in
    let c = Cursor.create ~txn dbs.tqueue in
    let run () = 
      let (timestr,keystr) = Cursor.get c Cursor.FIRST [] in
      Cursor.del c;
      (float_of_string timestr, key_of_string keystr)
    in
    protect ~f:run ~finally:(fun () -> Cursor.close c)
    

  (***********************************************************************)

  type key_metadata = { md_hash: string;
			md_words: string list;
			md_keyid: string;
			md_time: float;
			md_skey: skey;
		      }

  let shorten_offset offset = 
    if offset.pos <= Int64.of_int max_int then
      Offset { fnum = offset.fnum;
	       pos = Int64.to_int offset.pos;
	     }
    else
      LargeOffset offset

  let key_to_metadata_large_offset offset key = 
    { md_hash = KeyHash.hash key;
      md_words = key_extract_words key;
      md_keyid = Fingerprint.keyid_from_key ~short:true key;
      md_time = Unix.gettimeofday ();
      md_skey = shorten_offset offset;
    }
  let key_to_metadata_offset offset key = 
    { md_hash = KeyHash.hash key;
      md_words = key_extract_words key;
      md_keyid = Fingerprint.keyid_from_key ~short:true key;
      md_time = Unix.gettimeofday ();
      md_skey = Offset offset;
    }

  let key_to_metadata ?hash key = 
    { md_hash = (match hash with 
		   | None -> KeyHash.hash key
		   | Some hash -> hash);
      md_words = key_extract_words key;
      md_keyid = Fingerprint.keyid_from_key ~short:true key;
      md_time = Unix.gettimeofday ();
      md_skey = Key key;
    }

  (***********************************************************************)

  (** Bulk addition of key-metadata.  Used by fastbuild, so no transactional
    support provided.  *)
  let add_mds mds = 

    let dbs = get_dbs () in
    let mds = Array.of_list mds in

    (* Add hash-key mappings *)
    Array.sort mds ~cmp:(fun md1 md2 -> compare md1.md_hash md2.md_hash);
    Array.iter 
      ~f:(fun md -> 
	    try Db.put dbs.key ~key:md.md_hash 
	      ~data:(skey_to_string md.md_skey)
	      [Db.NOOVERWRITE]
	    with Key_exists -> ()
	 )
      mds;
    
    (* Add word-hash mappings *)
    let wordlist_array = 
      Array.map 
	~f:(fun md -> 
	      let words = md.md_words in
	      List.rev_map ~f:(fun word -> (word,md.md_hash)) words)
	mds
    in
    let words = flatten_array_of_lists wordlist_array in
    Array.sort ~cmp:compare words;
    Array.iter ~f:(fun (word,hash) -> 
		     try Db.put dbs.word ~key:word ~data:hash [Db.NODUPDATA]
		     with Key_exists -> ()
		  )
      words;

    (* Add keyid-hash mappings *)
    Array.sort mds ~cmp:(fun md1 md2 -> compare md1.md_keyid md2.md_keyid);
    Array.iter ~f:(fun md -> 
		     try Db.put dbs.keyid ~key:md.md_keyid 
		       ~data:md.md_hash [Db.NODUPDATA]
		     with Key_exists -> ())
      mds;

    (* Add time-hash mappings.  No sorting required *)
    Array.sort mds ~cmp:(fun md1 md2 -> compare md1.md_time md2.md_time);
    Array.iter mds
      ~f:(fun md -> 
	    let timestr = float_to_string md.md_time
	    and eventstr = event_to_string (Add md.md_hash) in
	    Db.put dbs.time ~key:timestr ~data:eventstr [Db.NODUPDATA])
      
      
  (****************************************************************)

  (** add a single key metadata with transaction possibly passed in *)
  let add_md_txn ?txn md = 
    let dbs = get_dbs () in
    
    Db.put ?txn dbs.key ~key:md.md_hash ~data:(skey_to_string md.md_skey) 
      [Db.NOOVERWRITE];
    
    List.iter md.md_words 
      ~f:(fun word -> Db.put ?txn dbs.word ~key:word ~data:md.md_hash [
	    Db.NODUPDATA] );
    
    Db.put ?txn dbs.keyid ~key:md.md_keyid ~data:md.md_hash [Db.NODUPDATA];
    let timestr = float_to_string (Unix.gettimeofday ()) 
    and eventstr = event_to_string (Add md.md_hash) in
    Db.put ?txn dbs.time ~key:timestr ~data:eventstr []
    
  (** add a single key with transaction possibly passed in *)
  let add_key_txn ?txn ?hash key = 
    let md = key_to_metadata ?hash key in
    add_md_txn ?txn md


  (** Does the required transactional wrapping around add_key_txn *)
  let add_key ?parent ?hash key = 
    let dbs = get_dbs () in
    let txn = 
      (if withtxn 
       then Some (Txn.txn_begin dbs.dbenv parent []) 
       else None) 
    in
    try
      add_key_txn ?txn ?hash key;
      (match txn with None  -> () | Some txn -> Txn.commit txn [])
    with
	e -> 
	  match txn with
	      None -> raise e
	    | Some txn -> 
		ignore (eplerror 2 e "Keydb.add_key -- Aborting transaction");
		Txn.abort txn;
		raise e


(****************************************************************)

  (** Does transactional wrapping around key adding, 
    allowing multiple keys to be added in a single transaction.*)
  let add_multi_keys keys = 
    let dbs = get_dbs () in
    let txn = 
      (if withtxn
       then Some (Txn.txn_begin dbs.dbenv None [])
       else None)
    in
    try
      List.iter 
	~f:(fun key -> 
	      try 
		add_key ?parent:txn key
	      with
		| Key_exists ->
		    ignore (plerror 2 "%s"
			      ("add_multi_keys: Key_exists. " ^
			       "continuing transaction"));
		    let hashstr = KeyHash.hexify (KeyHash.hash key) in
		    ignore (plerror 4 "Hash of duplicate key: %s" hashstr)
		| e ->
		    ignore (eplerror 2 e "%s" 
			      ("add_multi_keys: unexpected error.  " ^
			       "Continuing transaction on other keys"))
	   )
	keys;
      (match txn with None -> () | Some txn -> Txn.commit txn [])
    with
	e -> match txn with 
	    None -> raise e
	  | Some txn -> 
	      ignore (eplerror 2 e 
			"Keydb.add_multi_key -- Aborting transaction");
	      Txn.abort txn;
	      raise e
	      

  (***********************************************************************)

  (** Adds multiple keys at once *)
  let add_keys keys = 
    let mds = List.map ~f:key_to_metadata keys in
    add_mds mds

  (***********************************************************************)

  let add_key_merge_txn ?txn ?hash key = 
    let hash = match hash with 
	None -> KeyHash.hash key 
      | Some hash -> hash 
    in

    let dbs = get_dbs () in
    let keyid = Fingerprint.keyid_from_key ~short:true key in
    let keys = get_by_keyid keyid in

    (* Aborts transaction if it would have no result.  This way, 
       you don't get pointless database munging as well 
       as polution of the log. *)
    if (List.length keys = 1 
         && (KeyHash.hash (List.hd keys) = hash
	     (* || Set.subset (Set.of_list key) (Set.of_list (List.hd keys)) *)
	    )
       )
    then raise Key_exists;

    signore (perror "%d keys found with same fingerprint.  Computing merged keys" 
	       (List.length keys));
    let mergedkey = 
      List.fold_left 
	~f:(fun key oldkey -> 
	      match KeyMerge.merge key oldkey with
		  None -> 
		    signore (plerror 5 "Key merge failed.");
		    key
		| Some mergedkey ->
		    if KeyHash.hash mergedkey = KeyHash.hash oldkey
		    then (
		      signore (plerror 5 "Key merge resulted in no change");
		      mergedkey
		    ) else (
		      signore (plerror 5 "Key merge succeeded. Deleting old key.");
		      delete_key ?txn oldkey;
		      mergedkey
		    )
	   )
	~init:key keys
    in
    signore (perror "Calling add_key_txn");
    try 
      add_key_txn ?txn mergedkey
    with
	Key_exists ->
	  signore (plerror 3 "%s Hash: %s"
		     ("Key addition caused merger of existing keys " ^
		      "but no new key added. ")
		     (KeyHash.hexify hash))


  (**********************************************************)

  let add_key_merge ?(newkey=false) ?parent ?hash key = 
    let dbs = get_dbs () in
    let txn = 
      (if withtxn 
       then Some (Txn.txn_begin dbs.dbenv parent []) 
       else None) in
    try
      add_key_merge_txn ?txn ?hash key;
      if newkey then 
	(match txn with 
	     None -> failwith "txn required for enqueue_key"
	   | Some txn -> enqueue_key ~txn key);

      (match txn with None  -> () | Some txn -> Txn.commit txn [])
    with
	e -> 
	  match txn with
	      None -> raise e
	    | Some txn -> 
		ignore (eplerror 2 e 
			  "Keydb.add_key_merge -- Aborting transaction");
		Txn.abort txn;
		raise e


  (**********************************************************)

  (** Does transactional wrapping around key adding, allowing multiple keys
    to be added (with key merging) in a single transaction.*)
  let add_multi_keys_merge keys =
    let dbs = get_dbs () in
    let txn = 
      (if withtxn
       then Some (Txn.txn_begin dbs.dbenv None [])
       else None)
    in
    try
      List.iter 
	~f:(fun key -> 
	      try 
		add_key_merge ?parent:txn key
	      with
		  Key_exists ->
		    ignore (perror "%s" 
			      ("add_multi_keys_merge: Key_exists. " ^
			       "continuing transaction"));
		    let hashstr = KeyHash.hexify (KeyHash.hash key) in
		    ignore (perror "Hash of duplicate key: %s"
			      hashstr)
		| e ->
		    ignore (eplerror 2 e "%s" 
			      ("add_multi_keys_merge: unexpected error.  " ^
			       "Continuing transaction on other keys"))
		    
	   )
	keys;
      (match txn with None -> () | Some txn -> Txn.commit txn [])
    with
	e -> match txn with 
	    None -> raise e
	  | Some txn -> 
	      ignore (eplerror 2 e 
			"Keydb.add_multi_keys_merge -- Aborting transaction");
	      Txn.abort txn;
	      raise e


end
