(** Code for reading in and processing files received from 
  PKS-style email-based sync *)
open Common
open StdLabels
open MoreLabels
open Printf


let max_filesize = 200 * 1024
let input_msg f = 
  let b = Buffer.create (min max_filesize (in_channel_length f)) in
  Buffer.add_channel b f (in_channel_length f);
  Buffer.contents b


let dirname = "messages"

let lsdir dir = 
  let dirhandle = Unix.opendir dir in
  let run () = 
    let rec loop accum = 
      match (try Some (Unix.readdir dirhandle)
	     with End_of_file -> None)
      with
	  Some fname -> loop (fname::accum)
	| None -> accum
    in
    List.map ~f:(Filename.concat dir) (loop [])
  in
  protect ~f:run ~finally:(fun () -> Unix.closedir dirhandle)

(** reads specified mail file and returns key if any *)
let load_message fname = 
  let file = open_in fname in
  let run () = 
    let text = input_msg file in
    let msg = Recvmail.parse text in
    msg.Sendmail.body
  in
  protect ~f:run ~finally:(fun () -> close_in file)


let get_mtime fname = (Unix.stat fname).Unix.st_mtime

let demote fname = 
  if Sys.file_exists fname then
    let destdir = !Settings.failed_msgdir in
    if not (Sys.file_exists destdir) then
      Unix.mkdir destdir 0o700;
    Sys.rename fname (Filename.concat destdir (Filename.basename fname))

(****************************************************************************)
(* Event Handlers  **********************************************************)
(****************************************************************************)

(** read any mails in queue directory, process them, and remove them *)
let rec load_mailed_keys ~addkey () = 
  signore (plerror 5 "checking for key emails");
  let files = try lsdir !Settings.msgdir with Unix.Unix_error _ -> [] in
  let ready_files = 
    List.filter ~f:(fun file -> Filename.check_suffix file ".ready") files 
  in
  List.iter ready_files
    ~f:(fun fname -> 
	  try
	    (*if Unix.gettimeofday () -. get_mtime fname > (30. *. 60.)
	    then signore (plerror 3 "Ignoring file '%s'.  File too old." fname)
	    else *)
	    let text = load_message fname in
	    let keys = Armor.decode_pubkey text in
	    signore (plerror 3 "Adding list of %d keys from file %s"
		       (List.length keys) fname);
	    List.iter ~f:(fun key -> 
			    try addkey key
			    with Bdb.Key_exists -> ())
	      keys;
	    Sys.remove fname
	  with
	    | Eventloop.SigAlarm | Sys.Break as e -> raise e
	    | e -> 
		signore (eplerror 2 e "Failure adding key from file %s. %s" fname
			   "Moving to failed_messages.");
		demote fname
       );
  []


