(** Simple module for loading membership information.  Currently only loads
  membership from membership file.
  @author Yaron M. Minsky
*)
open StdLabels
open MoreLabels
module Unix=UnixLabels
open Printf
open Scanf
open Common

exception Bug of string

let membership = ref ([| |],-1.)

let whitespace = Str.regexp "[ \t]+"

let lookup_hostname string = 
  (Unix.gethostbyname string).Unix.h_addr_list.(0)

let local_recon_addr =  
  Unix.ADDR_INET (lookup_hostname !Settings.hostname, recon_port)
let remove_self addresses = 
  List.filter ~f:( (<>) local_recon_addr ) addresses

let load_membership_file file = 
  let m = ref [] in
  ( try
      while true do
	try
	  let l = input_line file in
	  let pieces = Str.split whitespace l in
	  let entry = 
	    sscanf l "%s %d" 
	      (fun addr port ->
		 Unix.ADDR_INET (lookup_hostname addr,port) )
	  in
	  m := entry::!m 
	with
	    End_of_file -> raise Exit
      done
    with
	Exit -> ()
  );
  Array.of_list (remove_self !m)


let get_mtime fname = (Unix.stat fname).Unix.st_mtime

let load_membership fname = 
  let file = open_in fname in
  protect ~f:(fun () -> 
		let mshp = load_membership_file file in
		membership := (mshp,get_mtime fname)
	     )
    ~finally:(fun () -> close_in file)

let sockaddr_to_string sockaddr = match sockaddr with
    Unix.ADDR_UNIX s -> sprintf "<ADDR_UNIX %s>" s
  | Unix.ADDR_INET (addr,p) -> sprintf "<ADDR_INET %s:%d>" 
      (Unix.string_of_inet_addr addr) p

let membership_string () = 
  let (mshp,_) = !membership in
  let strings = List.map ~f:sockaddr_to_string (Array.to_list mshp) in
  "Membership: " ^ String.concat ~sep:", " strings
    

let reload_if_changed () = 
  let fname = !Settings.membership_file in
  let (mshp,old_mtime) = !membership in
  if old_mtime <> get_mtime fname then 
    ( load_membership fname;
      ignore (plerror 5 "%s" (membership_string ()))
    )

let get () = 
  if Sys.file_exists !Settings.membership_file then (
    reload_if_changed ();
    let (m,mtime) = !membership in 
    m
  )
  else [| |]

let same_inet_addr addr1 addr2 = 
  match (addr1,addr2) with
      (Unix.ADDR_INET (ip1,_), Unix.ADDR_INET (ip2,_)) -> ip1 = ip2
    | _ -> false

let test addr = 
  reload_if_changed ();
  let (m,mtime) = !membership in
  
  let found = ref false in
  let i = ref 0 in
  while !i < Array.length m && not !found do 
    if same_inet_addr addr m.(!i) then
      found := true;
    incr i
  done;
  !found



(************************************************************)
(** Code for keeping track of hosts to send mail updates to *)
(************************************************************)

let mailsync_partners = ref ([ ],-1.)

let rec load_mailsync_partners_file file = 
  try 
    let email = Wserver.strip (input_line file) in
    if email.[0] <> '#' && String.contains email '@'
    then email::(load_mailsync_partners_file file)
    else load_mailsync_partners_file file
  with
      End_of_file -> []

let load_mailsync_partners fname = 
  let file = open_in fname in
  let run () = 
    mailsync_partners := (load_mailsync_partners_file file,
			  get_mtime fname)
  in
  protect ~f:run ~finally:(fun () -> close_in file)

let reload_mailsync_if_changed () = 
  let fname = !Settings.mailsync_file in
  let (mshp,old_mtime) = !membership in
  if old_mtime <> get_mtime fname then 
    load_mailsync_partners fname

let get_mailsync_partners () = 
  let partners = 
    if Sys.file_exists !Settings.membership_file then (
      reload_mailsync_if_changed ();
      let (m,mtime) = !mailsync_partners in
      m
    )
    else []
  in
  if partners = [] then failwith "No partners specified"
  else partners
