open StdLabels
open MoreLabels
open Bdb
open Common
module Unix = UnixLabels

module type PTreeSettings =
sig
  val mbar: int
  val bitquantum: int
  val treetype: [ `ondisk | `transactional | `inmem ]
  val max_nodes: int
  val dbdir: string
  val cache_bytes: int option
  val pagesize: int option
end

module Make(M: PTreeSettings) =
struct

  include M
  exception No_db

  let initialized = ref false

  type dbstate = { dbenv: Dbenv.t;
		   db: Db.t;
		 }

  let num_samples = mbar + 1

  (***************************************************************)
  (* Database and PTree setup ************************************)
  (***************************************************************)

  (** DB access methods.  The following will be passed on to the prefixtree *)

  let dbstate = ref None

  let closedb () = 
    match !dbstate with 
	None -> ()
      | Some dbs -> 
	  Db.close dbs.db;
	  Dbenv.close dbs.dbenv;
	  initialized := false

  let get_dbs () = 
    match !dbstate with None -> raise No_db
      | Some dbs -> dbs

  let load key = 
    let dbs = get_dbs () in
    let rval = Db.get dbs.db key [] in
    rval

  let save txn ~key ~data = 
    let dbs = get_dbs () in
    Db.put ?txn dbs.db ~key ~data []

  let delete txn key = 
    let dbs = get_dbs () in
    Db.del ?txn dbs.db key

  (** Returns a tuple containing database information needed by ptree *)
  let get_db () = match !dbstate with
      None -> None
    | Some dbs -> Some (load,save,delete,max_nodes)

  (*****************************************************************)
  (** txnopt operations do nothing if transactions are not enabled *)

  let new_txnopt () =  
    if treetype = `transactional then
      let dbs = get_dbs () in
      Some (Txn.txn_begin dbs.dbenv None [])
    else None 

  let commit_txnopt txn = 
    (match txn with None -> () | Some txn -> Txn.commit txn []) 

  let abort_txnopt txn = 
    (match txn with None -> () | Some txn -> Txn.abort txn)

  let checkpoint ?(kbyte=0) ?(min=0) () = 
    match !dbstate with
	None -> ()
      | Some dbs -> 
	  if treetype = `transactional then (
	    ignore (perror "Checkpointing database");
	    Txn.checkpoint dbs.dbenv ~kbyte ~min [];
	    ignore (perror "Checkpoint done")
	  )

  (*****************************************************************)

  (** Set up ptree database if such is necessary *)
  let open_ptree_db () = match treetype with

    | `inmem -> None

    | `ondisk | `transactional as treetype -> 
      ignore (plerror 3 "Opening PTree database");

      if not (Sys.file_exists dbdir )
      then Unix.mkdir dbdir ~perm:0o700;
      let dbenv = Dbenv.create [] in
      ( match cache_bytes with None -> ()
	  | Some cache_bytes -> Dbenv.set_cachesize dbenv
	      ~gbytes:0 ~bytes:cache_bytes ~ncache:0);
      Dbenv.dopen dbenv dbdir
	([Dbenv.INIT_MPOOL; Dbenv.CREATE] @ ( 
	   match treetype with
	    | `transactional -> [Dbenv.INIT_TXN; Dbenv.RECOVER]
	    | `ondisk -> []))
	0o600;
      let db = Db.create ~dbenv [] in
      ( match pagesize with 
	  | None -> ()
	  | Some pagesize -> Db.set_pagesize db pagesize );
      Db.dopen db "ptree" Db.BTREE [Db.CREATE] 0o600;
      Some { dbenv = dbenv;
	     db = db; 
	   }

  let init () = 
    if !initialized then failwith "Attempt to re-initialize PTreeDB";
    dbstate := open_ptree_db ();
    initialized := true

end
