open StdLabels
open MoreLabels

open Common
open Packet
open Printf

exception Partial_body_length of int

(********************************************************)

(** parse new-style packet length *)
let parse_new_packet_length cin = 
  let byte1 = cin#read_byte in
  if byte1 <= 191 then byte1  (* one-octet length *)
  else if byte1 <= 223  then (* two-octet length *) 
    let byte2 = cin#read_byte in 
    (byte1 - 192) lsl 8 + byte2 + 192
  else if byte1 = 255 then (* five-octet length *)
    let byte2 = cin#read_byte in 
    let byte3 = cin#read_byte in
    let byte4 = cin#read_byte in
    let byte5 = cin#read_byte in
    (byte2 lsl 24) lor (byte3 lsl 16) lor (byte4 lsl 8) lor byte5
  else (* partial body length *)
    raise (Partial_body_length (1 lsl (byte1 land 0x1f)))

(********************************************************)

let read_packet cin = 
  let packet_tag = cin#read_byte in
  if ((packet_tag lsr 7) land 1 <> 1) 
  then failwith (sprintf "Bit 7 of packet tag was not 1 as expected: %x" 
		   packet_tag);
  match (packet_tag lsr 6) land 1 with

      0 -> (* old format *)
	let content_tag = (packet_tag land 0b111100) lsr 2
	and length_type = packet_tag land 0b11
	in 
	(match length_type with
	     0 | 1 | 2 -> 
	       let length_length = 1 lsl length_type in
	       let length_str = cin#read_string length_length in
	       let length = Utils.int_from_bstring length_str 
			      ~pos:0 ~len:length_length in
	       { (* packet_tag = packet_tag; *)
		 content_tag = content_tag;
		 packet_type = content_tag_to_ptype content_tag;
		 packet_length = length;
		 packet_body = cin#read_string length;
	       }
	       
	   | 3 -> (* indeterminate length header --- extends to end of file *)
	       failwith "Unexpected indeterminate length packet"
	   | _ -> 
	       failwith "Unexpected length type"
	)

    | 1 -> (* new_format *)
	let content_tag = packet_tag land 0b111111 in
	let length = parse_new_packet_length cin in
	{ (* packet_tag = packet_tag; *)
	  content_tag = content_tag;
	  packet_type = content_tag_to_ptype content_tag;
	  packet_length = length;
	  packet_body = cin#read_string length;
	}

    | _ -> raise (Bug "ParsePGP.read_packet: expected 0/1 value")

    
(********************************************************)

let offset_read_packet cin = 
  let offset = LargeFile.pos_in cin#inchan in
  let packet = read_packet cin in
  (offset,packet)

(********************************************************)

let offset_length_read_packet cin = 
  let offset = pos_in cin#inchan in
  let packet = read_packet cin in
  let final_offset = pos_in cin#inchan in
  (packet,offset,final_offset - offset)

(********************************************************)

let read_mpi cin = 
  let length = cin#read_int_size 2 in
  let data = cin#read_string 
	       ((length + 7)/8)
  in
  { mpi_bits = length; mpi_data = data }

(********************************************************)

let parse_pubkey_info packet = 
  let cin = new Channel.string_in_channel packet.packet_body 0 in
  let version = cin#read_byte in
  let creation_time = cin#read_int32 in
  let (algorithm,mpi) = 
    match version with
      | 4 -> 
	  let algorithm = cin#read_byte in
	  let mpi = read_mpi cin in
	  (algorithm,mpi)
      | 2 | 3 ->
	  let expiration = cin#read_int_size 2 in
	  let algorithm = cin#read_byte in
	  let mpi = read_mpi cin in
	  (algorithm,mpi)
      | _ -> failwith (sprintf "Unexpected pubkey version: %d" version)
  in
  { pk_version = version;
    pk_ctime = creation_time;
    pk_alg = algorithm;
    pk_keylen = mpi.mpi_bits;
  }
  
(********************************************************)

let parse_siginfo packet = 
  let cin = new Channel.string_in_channel packet.packet_body 0 in
  let version = cin#read_byte in
  match version with
    | 2 | 3 ->
	cin#skip 1; (* length packet which must be 5 *)
	let sigtype = cin#read_byte in
	let ctime = cin#read_int in
	cin#skip 8;
	let pk_alg = cin#read_byte in
	let hash_alg = cin#read_byte in
	{ si_version = version;
	  si_sigtype = int_to_sigtype sigtype;
	  si_pk_alg = pk_alg;
	  si_hash_alg = hash_alg;
	}
    | 4 ->
	let sigtype = cin#read_byte in
	let pk_alg = cin#read_byte in
	let hash_alg = cin#read_byte in
	{ si_version = version;
	  si_sigtype = int_to_sigtype sigtype;
	  si_pk_alg = pk_alg;
	  si_hash_alg = hash_alg;
	}
    | _ -> failwith (sprintf "Unexpected signature version: %d" version)
	  

(** Parsing of signature subpackets *)

(** parse sigsubpacket length *)
let parse_sigsubpacket_length cin = 
  let byte1 = cin#read_byte in
  if byte1 < 192 then byte1 (* one octet length *)
  else if byte1  < 255 then
    let byte2 = cin#read_byte in
    ((byte1 - 192) lsl 8) + (byte2) + 192
  else if byte1 = 255 then (* five-octet length *)
    let byte2 = cin#read_byte in 
    let byte3 = cin#read_byte in
    let byte4 = cin#read_byte in
    let byte5 = cin#read_byte in
    (byte2 lsl 24) lor (byte3 lsl 16) lor (byte4 lsl 8) lor byte5
  else
    failwith "Unable to parse sigsubpacket length"
    
let read_sigsubpacket cin = 
  let length = parse_sigsubpacket_length cin in
  let ssp_type = cin#read_byte in
  let body = cin#read_string (length - 1) in
  { ssp_length = length;
    ssp_type = ssp_type;
    ssp_body = body;
  }

let rec has_nomodify_sigsubpacket cin = 
  try
    let packet = read_sigsubpacket cin in
    if packet.ssp_type = 23 && 
      (int_of_char packet.ssp_body.[0]) land 0x80 = 0x80 
    then true
    else has_nomodify_sigsubpacket cin
  with
      Channel.TooShort _ -> false

let get_hashed_subpacket_string cin = 
  let version = cin#read_byte in
  if version <> 4 then 
    failwith "Attempt to parse non-v4 signature as v4 signature";
  let sigtype = cin#read_byte in
  let key_alg = cin#read_byte in
  let hash_alg = cin#read_byte in
  let hashed_subpacket_count = cin#read_int_size 2 in
  (* now we can start reading the hashed sub-packets *)
  cin#read_string hashed_subpacket_count 

let signature_has_nomodify sigpacket = 
  let cin = new Channel.string_in_channel 
	      sigpacket.packet_body 0 in
  let hashed_subpacket_string = get_hashed_subpacket_string cin in
  let cin = new Channel.string_in_channel hashed_subpacket_string 0 in
  has_nomodify_sigsubpacket cin

  
  

