(** Executable:  Builds a prefix-tree database from an existing Keydb *)
open StdLabels
open MoreLabels
open Printf
open Common
open Bdb
module PTree = PrefixTree

module Keydb = Keydb.Make(struct 
			    let withtxn = false
			    and cache_bytes = !Settings.cache_bytes
			    and pagesize = !Settings.pagesize
			    and dbdir = !Settings.dbdir
			    and dumpdir = !Settings.dumpdir
			  end)

module PTreeDB = 
  PTreeDB.Make(struct
		 let mbar = !Settings.mbar 
		 and bitquantum = !Settings.bitquantum
		 and treetype = `ondisk
		 and max_nodes = !Settings.max_ptree_nodes
		 and dbdir = !Settings.ptree_dbdir
		 and cache_bytes = !Settings.ptree_cache_bytes
		 and pagesize = !Settings.ptree_pagesize
	       end)
open PTreeDB

let rec get_n n str = match n with
    0 -> []
  | _ -> 
      match SStream.next str with
	  None -> []
	| Some x -> x::(get_n (n-1) str)

let process_hashes hashes ptree =
  List.iter ~f:(PTree.insert_str ptree None) hashes

let run str () = 
  let ptree = PTree.create ?db:(get_db ()) ~txn:None 
		~num_samples ~bitquantum 
		~thresh:(mbar * !Settings.ptree_thresh_mult) ()
  in
  let count = ref 0 in
  while 
    match get_n 5000 str with
	[] -> false
      | hashes ->
	  process_hashes hashes ptree;
	  count := !count + List.length hashes;
	  ignore (perror "%d hashes processed" !count);
	  true
  do () done;
  let last_ts = Keydb.last_ts () in
  PTree.set_synctime ptree last_ts;
  ignore (perror "Cleaning Tree.");
  PTree.clean None ptree

let () = 
  if Sys.file_exists !Settings.ptree_dbdir then
    ignore (Unix.system (sprintf "rm -r %s" !Settings.ptree_dbdir));

  PTreeDB.init ();

  ignore (perror "Opening dbs...");
  Keydb.open_dbs ();
  
  let (hstr,hstr_close) = Keydb.create_hashstream () in
  protect ~f:(run hstr)
    ~finally:(fun () -> 
		PTreeDB.closedb ();
		hstr_close ();
		Keydb.close_dbs ();
	     )
