(** Reconciliation logic that is shared between the client and server *)
open StdLabels
open MoreLabels
open Common
open CMarshal
open ReconMessages
open Printf

(** Configuration related functions *)

(** Build map containing configuration information *)
let build_configdata () = 
  let map = Map.empty in
  let map = (map |< "version") version in
  let map = (map |< "http port") (int_to_string http_port) in
  let map = (map |< "bitquantum") (int_to_string !Settings.bitquantum) in
  let map = (map |< "mbar") (int_to_string !Settings.mbar) in
  map

(** Returns `passed if there are no problems with the configdata,
  `failed s if there is a problem, where s is a string describing the 
  problem.
*)
let test_configdata cd =
  try
    let remote_version_string =  cd |= "version" in
    let remote_version = parse_version_string remote_version_string in
    if remote_version < version_tuple
    then `failed (sprintf "Requires version at least %s.  %s provided " 
		    version remote_version_string)
    else
      let bitquantum = int_of_string (cd |= "bitquantum") in
      let mbar = int_of_string (cd |= "mbar") in
      if bitquantum <> !Settings.bitquantum then 
	`failed "bitquantum values do not match"
      else if mbar <> !Settings.mbar then
	`failed "mbar values do not match"
      else 
	`passed
  with
      Not_found -> `failed "Missing entry in configdata"
    | e -> 
	Eventloop.reraise e;
	`failed (sprintf "Error parsing configdata: %s" 
		   (Printexc.to_string e) )

(** Exchanges config data with other host, and tests 
  whether provided config data allows for reconciliation 
  to proceed. *)
let handle_config cin cout = 
  let configdata = build_configdata () in
  marshal cout (Config configdata); (* channel is flushed here *)
  let remote_configdata = 
    match (unmarshal cin).msg with
      | Config x -> x
      | _ -> failwith "No configdata provided"
  in
  (match test_configdata remote_configdata with
     | `passed -> 
	 marshal_string cout "passed"; 
	 cout#flush
     | `failed reason -> 
	 marshal_string cout "failed";
	 marshal_string cout reason;
	 cout#flush;
	 failwith (sprintf "Remote configuration rejected: %s"
		     reason)
  );
  (match unmarshal_string cin with
       "passed" -> ()
     | "failed" ->
	 let reason = unmarshal_string cin in
	 failwith (sprintf "Local configuration rejected remotely: %s"
		     reason)
     | _ -> failwith "Unexpected configuration confirmation response"
  );
  remote_configdata


let config_get_http_port cd = 
  int_of_string (cd |= "http port")

let change_port sockaddr newport = match sockaddr with
    Unix.ADDR_UNIX _ -> raise (Invalid_argument 
				 "Can't change port of UNIX address")
  | Unix.ADDR_INET (ipaddr,port) -> Unix.ADDR_INET (ipaddr,newport)

let print_config config = 
  signore (perror "Printing config");
  Map.iter ~f:(fun ~key ~data ->
		 signore (perror "   %s: %s" key data)
	      )
    config
