(** Executable: server process that handles reconciliation. *)

open StdLabels
open MoreLabels
open Printf
open Common
open Packet
module Unix = UnixLabels
module PTree = PrefixTree
open DbMessages

module Map = PMap.Map

let () = set_logfile ".recon"
(* let () = Gc.set { (Gc.get()) with Gc.verbose = 0x3FF }   *)

module PTreeDB = 
  PTreeDB.Make(struct
		 let mbar = !Settings.mbar 
		 and bitquantum = !Settings.bitquantum
		 and treetype = 
		   if !Settings.transactions
		   then `transactional
		   else if !Settings.disk_ptree 
		   then `ondisk else `inmem
		 and max_nodes = !Settings.max_ptree_nodes
		 and dbdir = !Settings.ptree_dbdir
		 and cache_bytes = !Settings.ptree_cache_bytes
		 and pagesize = !Settings.ptree_pagesize
	       end)
open PTreeDB

let () = 
  signore (plerror 1 "sks_recon, SKS version %s" version); 
  signore (plerror 1 "Copyright Yaron Minsky 2002"); 
  signore (plerror 1 "Licensed under GPL.  See COPYING file for details"); 
  signore (plerror 5 "recon port: %d" recon_port)

let hash_bundle_size = !Settings.http_fetch_size
let recon_addr = Unix.ADDR_INET (Unix.inet_addr_any,recon_port)
let reconsock = Eventloop.create_sock recon_addr

let () = 
  if Sys.file_exists recon_command_name
  then Unix.unlink recon_command_name
let comsock = Eventloop.create_sock recon_command_addr

(******************************************************************)

(** Setup prefix tree, using disk-based access and transactions 
  as specified *)
let open_ptree () = 
  ignore (plerror 3 "Setting up PTree data structure");
  let txn = new_txnopt () in
  try 
    let ptree = 
      PTree.create ?db:(get_db ()) 
	~txn ~num_samples ~bitquantum 
	~thresh:(mbar * !Settings.ptree_thresh_mult)
	()
    in
    commit_txnopt txn;
    ptree
  with 
      e -> 
	abort_txnopt txn;
	closedb (); 
	raise e

(***************************************************************)

let () = PTreeDB.init ()
let ptree = ref (open_ptree ())

(** resets ptree to empty state or whatever is available on disk.  
  Called when there is reason to think the prefix-tree database 
  may be corrupted.
*)
let reopen_ptree () = 
  ignore (plerror 2 "Restarting PTree");
  (try closedb ()
   with e -> ignore (eplerror 1 e "Error closing database"));
  dbstate := open_ptree_db ();
  ptree := open_ptree ()

(** Evaluates f, and if f throws an exception, resets the ptree to a safe
  state.  f should be some function updating the prefixtree an exception 
  during which would indicate a possible corruption of the tree *)
let ptree_protect f =
  let txn = new_txnopt () in
  try 
    let rval = f txn in
    commit_txnopt txn;
    rval
  with
      Sys.Break ->
	abort_txnopt txn;
	raise Sys.Break
    | e -> 
	(* FIX: invalidate, don't reload until someone 
	   tries to gain access again *)
	abort_txnopt txn;
	ignore (eplerror 1 e "%s"
		  ("PTree may be corrupted.  " ^
		   "Restarting from scratch or from disk " ^
		   "as appropriate."));
	reopen_ptree ();
	raise e

(** Queue of bundles of hashes to be recovered*)
type recover_element = string list * Unix.sockaddr

let recover_list = Queue.create () 
let gossip_disabled () = not (Queue.is_empty recover_list)

(***************************************************************)
(**  Updating recover list *)
(***************************************************************)

let rec n_split list n = match (n,list) with
    (0,_) | (_,[]) -> ([],list) 
  | (_,hd::tl) -> 
      let (first,rest) = n_split tl (n-1) in
      (hd::first,rest)

      
let rec size_split list size = 
  match n_split list size with
      ([],[]) -> []
    | (first,rest) -> 
	let pieces = size_split rest size in
	first::pieces

let print_hashes hashes  =
  if List.length hashes = 0 then signore (plerror 4 "No hashes recovered")
  else if List.length hashes <= 10 then (
    ignore (plerror 4 "Hashes recovered: ");
    List.iter hashes
      ~f:(fun hash -> ignore (plerror 4 "\t%s" (KeyHash.hexify hash)));
  ) else
    ignore (plerror 4 "%d hashes recovered" (List.length hashes))

(** converts a list of elements of ZZp to a sorted list of hashes *)
let hashconvert elements =
  let hashes = List.map ~f:ZZp.to_bytes elements in
  let hashes = List.map ~f:(fun hash -> RMisc.truncate hash KeyHash.hash_bytes)
		 hashes in
  let hashes = List.sort ~cmp:compare hashes in
  hashes

(** Dumps the hashes associated with the difference set to the named file *)
let log_diffs log_fname diffs = 
  let hashes = hashconvert diffs in
  let hash_strs = List.map ~f:KeyHash.hexify hashes in
  let file = open_out log_fname in
  protect ~f:(fun () -> List.iter ~f:(fprintf file "%s\n") hash_strs)
    ~finally:(fun () -> close_out file)

let update_recover_list results partner_http_addr  =
  let hashes = hashconvert results in
  let bundles = size_split hashes hash_bundle_size in
  List.iter bundles ~f:(fun bundle -> 
			  Queue.add (bundle,partner_http_addr) 
			  recover_list)


(***************************************************************)
(**  Message Sending Primitives  *)
(***************************************************************)


(** send DbMessages message and wait for response *)
let send_dbmsg msg = 
  let ctr = ref 0 in
  let s = Unix.socket 
	    ~domain:Unix.PF_UNIX 
	    ~kind:Unix.SOCK_STREAM 
	    ~protocol:0 in
  protect ~f:(fun () ->
		Unix.connect s ~addr:db_command_addr;
		let cin = Channel.sys_in_from_fd s in
		let cout = Channel.sys_out_from_fd s in
		marshal cout msg;
		let reply = (unmarshal cin).msg in
		reply
	     )
    ~finally:(fun () -> Unix.close s)
    

(** send DbMessages message, don't wait for response *)
let send_dbmsg_noreply msg = 
  let s = Unix.socket 
	    ~domain:Unix.PF_UNIX
	    ~kind:Unix.SOCK_STREAM 
	    ~protocol:0 in
  protect ~f:(fun () ->
		Unix.connect s ~addr:db_command_addr;
		let cin = Channel.sys_in_from_fd s
		and cout = Channel.sys_out_from_fd s in
		marshal cout msg )
    ~finally:(fun () -> Unix.close s)

let is_content_type line = 
  try
    let colonpos = String.index line ':' in
    let prefix = String.sub ~pos:0 ~len:colonpos line in
    String.lowercase prefix = "content-type"
  with
      Not_found -> false

let get_keystrings_via_http addr hashes = 
  let s = Unix.socket 
	    ~domain:Unix.PF_INET 
	    ~kind:Unix.SOCK_STREAM 
	    ~protocol:0  in
  protect ~f:(fun () -> 
		let () = Unix.connect s ~addr in
		let cin = Channel.sys_in_from_fd s 
		and cout = Channel.sys_out_from_fd s in

		let sout = Channel.new_buffer_outc 0 in
		CMarshal.marshal_list ~f:CMarshal.marshal_string sout hashes;
		let msg = sout#contents in
		cout#write_string "POST /pks/hashquery\r\n";
		cout#write_string (sprintf "content-length: %d\r\n\r\n" 
				     (String.length msg));
		cout#write_string msg;
		cout#flush;
		ignore (input_line cin#inchan); (* read "HTTP" line *)
		let headers = Wserver.parse_headers Map.empty cin#inchan in
		let keystrings = 
		  CMarshal.unmarshal_list ~f:CMarshal.unmarshal_string cin
		in
		keystrings
	     )
    ~finally:(fun () -> Unix.close s)

(***************************************************************)
(*  Catchup Code   *********************************************)
(***************************************************************)

let rec last_ts log = match log with
    [] -> raise Not_found
  | (ts,event)::[] -> ts
  | hd::tl -> last_ts tl

let rec applylog txn log = match log with
    [] -> ()
  | (ts,Add hash)::tl ->
      PTree.set_synctime !ptree ts;
      PTree.insert_str !ptree txn hash;
      applylog txn tl
  | (ts,Delete hash)::tl ->
      PTree.set_synctime !ptree ts;
      PTree.delete_str !ptree txn hash;
      applylog txn tl

let rec catchup_rec count = 
  let resp = send_dbmsg (LogQuery (count,PTree.get_synctime !ptree)) in
  let log = 
    match resp with
	LogResp log -> log
      | _ -> failwith "Unexpected response"
  in
  match log with
      [] -> ()
    | _ -> 
	let length = List.length log in
	let newts = last_ts log in
	ptree_protect (fun txn -> 
			 applylog txn log;
			 ignore (plerror 5 "Added %d hash-updates. Caught up to %f" 
				   length newts));
	checkpoint ~kbyte:(1024 * 30) ~min:10 ();
	catchup_rec count 

let initial_catchup = ref true
let catchup () = 
  let count = 5000 in  (* Number to fetch at once *)
  catchup_rec count
  
let catchup_interval = 5.

(***************************************************************)
(*  Handlers  *************************************************)
(***************************************************************)

let eventify_handler handle = 
  (fun addr cin cout ->
     let cin = (new Channel.sys_in_channel cin)
     and cout = (new Channel.sys_out_channel cout) in
     handle addr cin cout
  )

let choose array = 
  if Array.length array = 0 then raise Not_found
  else array.(Random.int (Array.length array))

let choose_partner () = 
  try
    choose (Membership.get ())
  with
      Not_found | Invalid_argument _ -> 
	failwith "No gossip partners available"


(** The following three events are defined mutually recursively so they
  can schedule each other *)

let missing_keys_timeout = !Settings.missing_keys_timeout

let rec get_missing_keys () = 
  let cb = get_missing_keys in
  let name = "get missing keys" in
  let timeout = missing_keys_timeout in 
  try

    ( try
	let (hashes,httpaddr) = Queue.pop recover_list in
	ignore (plerror 3
		  "Requesting %d missing keys from %s, starting with %s"
		  (List.length hashes) (sockaddr_to_string httpaddr)
		  (match hashes with
		       [] -> "<nohash>"
		     | hash::tl -> KeyHash.hexify hash
		  )
	       );

	let keystrings = get_keystrings_via_http httpaddr hashes in
	ignore (plerror 3 "Keys received");
	let ack = send_dbmsg (KeyStrings keystrings) in
	if ack <> Ack 0 
	then failwith ("Reconserver.get_missing_keys: " ^
		       "Unexpected reply to KeyStrings message");
	[Eventloop.Event 
	   (Unix.gettimeofday (), 
	    Eventloop.make_tc ~name ~timeout ~cb; )
	]
      with
	| Queue.Empty -> 
	    [Eventloop.Event (Unix.gettimeofday (),
			      Eventloop.Callback 
				(fun () -> catchup (); [] ))
	    ]
	| Eventloop.SigAlarm as e -> raise e
	| e ->
	    Eventloop.reraise e;
	    signore (eperror e "Error getting missing keys");
	    [Eventloop.Event (Unix.gettimeofday (), 
			      Eventloop.make_tc ~cb ~timeout ~name)
	    ]
	    
    )
  with
    | Eventloop.SigAlarm ->
	(* If we time out, just schedule the next one *)
	[Eventloop.Event (Unix.gettimeofday (), 
			  Eventloop.make_tc ~cb ~timeout ~name; ) ]

	  

let sockaddr_to_name sockaddr = match sockaddr with
    Unix.ADDR_UNIX s -> sprintf "UNIX_%s" s
  | Unix.ADDR_INET (addr,p) -> sprintf "%s" (Unix.string_of_inet_addr addr)


let recon_handler addr cin cout = 
  if gossip_disabled ()  then (
    signore (plerror 3 
	       "Reconciliation attempt from %s while gossip disabled. %s"
	       (sockaddr_to_string addr) "Ignoring."
	    );
    []

  ) else if not (Membership.test addr) then (
    signore (plerror 1 
	       "Reconciliation attempt from unauthorized host %s.  Ignoring" 
	       (sockaddr_to_string addr) );
    []
  ) else (
    ignore (plerror 4 "Beginning recon as server, client: %s" 
	      (sockaddr_to_string addr)
	   );
    let cin = (new Channel.sys_in_channel cin)
    and cout = (new Channel.sys_out_channel cout) in
    let results = Server.handle !ptree addr cin cout in
    ignore (plerror 4 "Reconciliation complete");
    let (elements,partner) = results in
    let elements = Set.elements elements in
    print_hashes (hashconvert elements);
    if List.length elements > 0 
    then (
      log_diffs (sprintf "diff-%s.txt" (sockaddr_to_name addr)) elements;
      update_recover_list elements partner;
      [Eventloop.Event (Unix.gettimeofday () +. 10.0,
			Eventloop.make_tc ~cb:get_missing_keys
			  ~timeout:missing_keys_timeout
			  ~name:"get missing keys"
		       )]
    ) else 
      []
  ) 


let gossip () = if gossip_disabled () then [] else
  begin
    let partner = choose_partner () in
    ignore (plerror 4 "Recon partner: %s" 
	      (sockaddr_to_string partner));
    let (results,http_addr) = 
      Client.run_reconciliation !ptree ~partner 
	~self:Membership.local_recon_addr in
    let results = Set.elements results in
    ignore (plerror 4 "Reconciliation complete");
    print_hashes (hashconvert results);
    match results with
	[] -> []
      | _ ->
	  log_diffs (sprintf "diff-%s.txt"
		       (sockaddr_to_name partner)) 
	  results;
	  update_recover_list results http_addr;
	  [Eventloop.Event (Unix.gettimeofday (), 
			    Eventloop.make_tc ~cb:get_missing_keys
			   ~timeout:missing_keys_timeout
			   ~name:"get missing keys"
			   )]
  end

let command_handler addr cin cout = 
  match (unmarshal cin).msg with

    | Synchronize -> 
	marshal cout (Ack 0);
	ignore (plerror 2 "Initiating gossip due to explicit request");
	gossip ()

    | RandomDrop n -> 
	marshal cout (Ack 0);
	for i = 1 to n do
	  try
	    let hash = PTree.get_random !ptree (PTree.root !ptree) in
	    let hash = RMisc.truncate hash KeyHash.hash_bytes in
	    ignore (plerror 3 "Requesting deletion %s" 
		      (Utils.hexstring hash));
	    ignore (send_dbmsg (DeleteKey hash))
	  with
	      Not_found -> 
		failwith "Attempted to delete element from empty prefix tree"
	    | e -> 
		Eventloop.reraise e;
		ignore (eplerror 3 e "Attempt to delete key failed")
	done;
	[]

    | HashRequest hashes ->
	let keyresp = (send_dbmsg (HashRequest hashes)) in
	assert (match keyresp with Keys _ -> true | _ -> false);
	marshal cout keyresp;
	[]

    | Config (s,cvar) ->
	ignore (plerror 4 "Received config message");
	(match (s,cvar) with
	     ("maxnodes",`int x) -> 
	       ignore (plerror 3 "Setting maxnodes to %d" x);
	       let txn = new_txnopt () in
	       (try 
		  PTree.set_maxnodes !ptree txn x;
		  commit_txnopt txn
		with
		    e -> 
		      ignore (eplerror 1 e 
				"set_maxnodes Transaction aborting");
		      abort_txnopt txn)
	   | _ -> 
	       failwith "Unexpected config request"
	);
	[]
	      
    | m -> 
	marshal cout ProtocolError;
	ignore (perror "Unexpected message: %s" (msg_to_string m));
	[]

(***************************************************************)

let sync_interval = !Settings.recon_sync_interval
let sync_tree () = 
  ignore (perror "Syncing prefix tree");
  let txn = new_txnopt () in
  try
    PTree.clean txn !ptree;
    commit_txnopt txn
  with
      e -> 
	ignore (eplerror 1 e "sync_tree transaction aborting");
	abort_txnopt txn;
	raise e
	

let checkpoint_interval = !Settings.recon_checkpoint_interval

(***************************************************************)


let run () = 
  Eventloop.evloop

    ( [ Eventloop.Event (0.0, Eventloop.Callback (fun () -> catchup (); [])) ]
      @ (Ehandlers.repeat_forever_simple catchup_interval catchup)
      @ (if !Settings.gossip 
	 then Ehandlers.repeat_forever !Settings.gossip_interval 
	   (Eventloop.make_tc ~cb:gossip ~name:"gossip"
	      ~timeout:!Settings.reconciliation_timeout 
	   )
	 else [] )
      @ (match treetype with
	   | `transactional -> 
	       Ehandlers.repeat_forever_simple checkpoint_interval checkpoint
	   | `ondisk -> Ehandlers.repeat_forever_simple sync_interval sync_tree
	   | `inmem -> []
	)
    )

    [ (comsock, Eventloop.make_th ~name:"command handler"
	 ~cb:(eventify_handler command_handler)
	 ~timeout:!Settings.command_timeout
      ); 
      (reconsock, Eventloop.make_th ~name:"reconciliation handler"
	 ~cb:recon_handler
	 ~timeout:!Settings.reconciliation_timeout
      ); 
    ]


let () = 
  protect ~f:run 
    ~finally:(fun () -> 
		let txn = new_txnopt () in
		try
		  PTree.clean txn !ptree;
		  commit_txnopt txn;
		  signore (plerror 2 "Tree cleaned");
		with
		    e -> 
		      signore (eplerror 1 e 
				 "Abort of final cleanup transaction");
		      abort_txnopt txn;
		      raise e
		      
		closedb ();
		signore (plerror 2 "DB closed")
	     )

